from pathlib import Path

import click
from colorama import Fore, Style

from vantage6.common import (
    bytes_to_base64s,
    debug,
    error,
    info,
    warning,
)
from vantage6.common.encryption import RSACryptor
from vantage6.common.globals import InstanceType

from vantage6.cli.common.utils import extract_name_and_is_sandbox
from vantage6.cli.configuration_create import select_configuration_questionnaire
from vantage6.cli.context.node import NodeContext
from vantage6.cli.globals import DEFAULT_NODE_SYSTEM_FOLDERS as N_FOL
from vantage6.cli.node.common import create_client_and_authenticate


@click.command()
@click.option("-n", "--name", default=None, help="Configuration name")
@click.option(
    "-c",
    "--config",
    default=None,
    help="Absolute path to configuration-file; overrides NAME",
)
@click.option(
    "--system",
    "system_folders",
    flag_value=True,
    help="Search for configuration in system folders rather than user folders",
)
@click.option(
    "--user",
    "system_folders",
    flag_value=False,
    default=N_FOL,
    help="Search for configuration in user folders rather than "
    "system folders. This is the default",
)
@click.option(
    "--no-upload",
    "upload",
    flag_value=False,
    default=True,
    help="Don't upload the public key to HQ",
)
@click.option(
    "-o",
    "--organization-name",
    default=None,
    help="Organization name. Used in the filename of the private key"
    " so that it can easily be recognized again later",
)
@click.option(
    "--overwrite",
    "overwrite",
    flag_value=True,
    default=False,
    help="Overwrite existing private key if present",
)
@click.option("--sandbox/--no-sandbox", "is_sandbox", default=False)
def cli_node_create_private_key(
    name: str,
    config: str,
    system_folders: bool,
    upload: bool,
    organization_name: str | None,
    overwrite: bool,
    is_sandbox: bool,
) -> None:
    """
    Create and upload a new private key

    Use this command with caution! Uploading a new key has several
    consequences, e.g. you and other users of your organization
    will no longer be able to read the results of tasks encrypted with current
    key.
    """
    name, is_sandbox = extract_name_and_is_sandbox(name, is_sandbox)
    if is_sandbox:
        system_folders = False

    NodeContext.LOGGING_ENABLED = False
    if config:
        name = Path(config).stem
        ctx = NodeContext(
            name,
            system_folders=system_folders,
            is_sandbox=is_sandbox,
            config_file=config,
        )
    elif not name:
        try:
            name = select_configuration_questionnaire(
                InstanceType.NODE, system_folders, is_sandbox
            )
        except Exception:
            error("No configurations could be found!")
            exit(1)
    ctx = NodeContext(name, system_folders=system_folders, is_sandbox=is_sandbox)

    # Authenticate with the hq to obtain organization name if it wasn't
    # provided
    if organization_name is None:
        client = create_client_and_authenticate(ctx, use_sandbox_port=ctx.is_sandbox)
        organization_name = client.whoami.organization_name

    # create directory where private key goes if it doesn't exist yet
    ctx.type_data_folder(system_folders).mkdir(parents=True, exist_ok=True)

    # generate new key, and save it
    filename = f"privkey_{organization_name}.pem"
    file_ = ctx.type_data_folder(system_folders) / filename

    if file_.exists():
        warning(f"File '{Fore.CYAN}{file_}{Style.RESET_ALL}' exists!")

        if overwrite:
            warning("'--overwrite' specified, so it will be overwritten ...")

    if file_.exists() and not overwrite:
        error("Could not create private key!")
        warning(
            "If you're **sure** you want to create a new key, "
            "please run this command with the '--overwrite' flag"
        )
        warning("Continuing with existing key instead!")
        private_key = RSACryptor(file_).private_key

    else:
        try:
            info("Generating new private key")
            private_key = RSACryptor.create_new_rsa_key(file_)

        except Exception as e:
            error(f"Could not create new private key '{file_}'!?")
            debug(e)
            info("Bailing out ...")
            exit(1)

        warning(f"Private key written to '{file_}'")
        warning(
            "If you're running multiple nodes, be sure to copy the private "
            "key to the appropriate directories!"
        )

    # create public key
    info("Deriving public key")
    public_key = RSACryptor.create_public_key_bytes(private_key)

    # update config file
    info("Updating configuration")
    ctx.config["node"]["encryption"]["private_key"] = str(file_)
    ctx.config_manager.put(ctx.config)
    ctx.config_manager.save(ctx.config_file)

    # upload key to the hq
    if upload:
        info(
            "Uploading public key to HQ. This will overwrite previously existing keys!"
        )

        if "client" not in locals():
            client = create_client_and_authenticate(
                ctx, use_sandbox_port=ctx.is_sandbox
            )

        # TODO what happens if the user doesn't have permission to upload key?
        # Does that lead to an exception or not?
        try:
            client.request(
                f"/organization/{client.whoami.organization_id}",
                method="patch",
                json={"public_key": bytes_to_base64s(public_key)},
            )

        except Exception as e:
            error("Could not upload the public key!")
            debug(e)
            exit(1)
    else:
        warning("Public key not uploaded!")

    info("[Done]")



from __future__ import annotations

from vantage6.common.context import AppContext

from vantage6.cli.globals import (
    DEFAULT_API_SERVICE_SYSTEM_FOLDERS as S_FOL,
    BackendType,
)


class BaseBackendContext(AppContext):
    """
    Base context for a vantage6 HQ or algorithm store

    Contains functions that the HQContext and AlgorithmStoreContext have
    in common.
    """

    def get_database_uri(self) -> str:
        """
        Obtain the database uri from the environment or the configuration.

        Returns
        -------
        str
            string representation of the database uri
        """
        # within container, the uri is set in the config file
        if self.in_container:
            return self.config.get("uri")

        # outside container, the uri is set in the values.yaml in a different location
        is_db_external = self.config.get("database", {}).get("external", False)
        if is_db_external:
            uri = self.config["database"]["uri"]
        else:
            db_conf = self.config["database"]
            uri = (
                f"postgresql://{db_conf['username']}:{db_conf['password']}@"
                f"{self.helm_release_name}-postgres:5432/{db_conf['name']}"
            )

        return uri

    @classmethod
    def from_external_config_file(
        cls,
        path: str,
        backend_type: BackendType,
        system_folders: bool = S_FOL,
        in_container: bool = False,
    ) -> BaseBackendContext:
        """
        Create a backend context from an external configuration file. External
        means that the configuration file is not located in the default folders
        but its location is specified by the user.

        Parameters
        ----------
        path : str
            Path of the configuration file
        backend_type : BackendType
            Type of backend, either 'hq' or 'algorithm-store'
        system_folders : bool, optional
            System wide or user configuration, by default S_FOL
        in_container : bool, optional
            Whether the application is running inside a container, by default False

        Returns
        -------
        HQContext | AlgorithmStoreContext
            Base backend context object
        """
        cls = super().from_external_config_file(
            path, backend_type, system_folders, in_container
        )
        return cls



import click

from vantage6.common import info
from vantage6.common.globals import InstanceType

from vantage6.cli.common.stop import execute_stop, helm_uninstall
from vantage6.cli.globals import DEFAULT_API_SERVICE_SYSTEM_FOLDERS, InfraComponentName
from vantage6.cli.k8s_config import KubernetesConfig


@click.command()
@click.option("-n", "--name", default=None, help="Configuration name")
@click.option("--context", default=None, help="Kubernetes context to use")
@click.option("--namespace", default=None, help="Kubernetes namespace to use")
@click.option(
    "--system",
    "system_folders",
    flag_value=True,
    default=DEFAULT_API_SERVICE_SYSTEM_FOLDERS,
    help="Search for configuration in system folders instead of user folders. "
    "This is the default.",
)
@click.option(
    "--user",
    "system_folders",
    flag_value=False,
    help="Search for configuration in the user folders instead of system folders.",
)
@click.option("--sandbox/--no-sandbox", "sandbox", default=False)
def cli_auth_stop(
    name: str,
    context: str,
    namespace: str,
    system_folders: bool,
    sandbox: bool,
):
    """
    Stop a running auth service.
    """
    execute_stop(
        stop_function=_stop_auth,
        instance_type=InstanceType.AUTH,
        infra_component=InfraComponentName.AUTH,
        stop_all=False,
        to_stop=name,
        namespace=namespace,
        context=context,
        system_folders=system_folders,
        is_sandbox=sandbox,
    )


def _stop_auth(auth_name: str, k8s_config: KubernetesConfig) -> None:
    info(f"Stopping auth {auth_name}...")

    # uninstall the helm release
    helm_uninstall(
        release_name=auth_name,
        k8s_config=k8s_config,
    )

    info(f"Auth {auth_name} stopped successfully.")



import sys

import click

from vantage6.common.globals import Ports

from vantage6.client import UserClient
from vantage6.client.utils import LogLevel

from vantage6.cli.test.common.diagnostic_runner import DiagnosticRunner
from vantage6.cli.utils import error


@click.command()
@click.option(
    "--hq-url",
    type=str,
    default=f"http://localhost:{Ports.SANDBOX_HQ}/hq",
    help="URL of HQ",
)
@click.option(
    "--auth-url",
    type=str,
    default=f"http://localhost:{Ports.SANDBOX_AUTH}",
    help="URL of the authentication service (Keycloak)",
)
@click.option(
    "--auth-realm",
    type=str,
    default="vantage6",
    help="Realm of the authentication service (Keycloak)",
)
@click.option(
    "--auth-client",
    type=str,
    default="public_client",
    help="Client ID of the authentication service (Keycloak)",
)
@click.option(
    "--collaboration",
    type=int,
    default=1,
    help="ID of the collaboration to create the task in",
)
@click.option(
    "-o",
    "--organizations",
    type=int,
    default=[],
    multiple=True,
    help="ID(s) of the organization(s) to create the task for",
)
@click.option(
    "--all-nodes",
    is_flag=True,
    help="Run the diagnostic test on all nodes in the collaboration",
)
@click.option(
    "--online-only",
    is_flag=True,
    help="Run the diagnostic test on only nodes that are online",
)
@click.option(
    "--private-key",
    type=str,
    default=None,
    help="Path to the private key for end-to-end encryption",
)
@click.option(
    "--session-id",
    type=int,
    default=1,
    help="ID of the session to use for the diagnostic test",
)
@click.option(
    "--database-label",
    type=str,
    default="olympic-athletes",
    help="Label of the database to use for the diagnostic test",
)
def cli_test_features(
    hq_url: str,
    auth_url: str,
    auth_realm: str,
    auth_client: str,
    collaboration: int,
    organizations: list[int] | None,
    all_nodes: bool,
    online_only: bool,
    private_key: str | None,
    session_id: int,
    database_label: str,
) -> list[dict]:
    """
    Run diagnostic checks on an existing vantage6 network.

    This command will create a task in the requested collaboration that will
    test the functionality of vantage6, and will report back the results.
    """
    if all_nodes and organizations:
        error("Cannot use --all-nodes and --organizations at the same time.")
        sys.exit(1)

    if all_nodes or not organizations:
        organizations = None

    client = UserClient(
        hq_url=hq_url,
        auth_url=auth_url,
        auth_realm=auth_realm,
        auth_client=auth_client,
        log_level=LogLevel.CRITICAL,
    )
    client.authenticate()
    client.setup_encryption(private_key)
    diagnose = DiagnosticRunner(
        client,
        collaboration,
        organizations,
        online_only,
        session_id,
        database_label,
    )
    res = diagnose()
    return res



"""
Development script to connect HQ to the local store.
"""

import time
from http import HTTPStatus

import requests

from vantage6.common import error, info
from vantage6.common.enum import AlgorithmStepType
from vantage6.common.globals import Ports

from vantage6.client import Client


def _wait_for_store_to_be_online(
    local_store_url: str, local_store_api_path: str
) -> None:
    """
    Wait for the store to be online.

    Parameters
    ---------
    client: Client
        The client to use to connect to vantage6 hub.
    local_store_url: str
        The URL of the local store.
    local_store_api_path: str
        The API path of the local store.
    """
    info(
        f"Waiting for store to be online at {local_store_url}{local_store_api_path}..."
    )
    max_retries = 100
    wait_time = 3
    ready = False
    print_unexpected_error = True
    for _ in range(max_retries):
        try:
            result = requests.get(f"{local_store_url}{local_store_api_path}/version")
            if result.status_code == HTTPStatus.OK:
                ready = True
                break
            elif print_unexpected_error:
                try:
                    error(f"Store returns unexpected error: {result.json()['msg']}")
                except Exception:
                    error(f"Store returns unexpected error: {result.status_code}")
                print_unexpected_error = False
        except Exception:
            info(f"Store not ready yet, waiting {wait_time} seconds...")
        time.sleep(wait_time)

    if not ready:
        error("Store did not become ready in time. Exiting...")
        exit(1)
    else:
        info("Store is online!")


def connect_store(client: Client, store_port: int = Ports.DEV_ALGO_STORE.value) -> str:
    """
    Connect HQ to the local store.

    Parameters
    ---------
    client: Client
        The client to use to connect to vantage6 hub.
    store_port: int
        The port of the store to connect to.

    Returns
    -------
    str
        A summary of the store connection process.
    """

    existing_stores = client.store.list().get("data", [])
    summary = "=== Store Connection Summary ===\n"

    # URL should be retrieved from the store, see issue:
    # https://github.com/vantage6/vantage6/issues/1824
    local_store_url = f"http://localhost:{store_port}"
    local_store_api_path = "/store"
    client.store.store_id = 1

    _wait_for_store_to_be_online(local_store_url, local_store_api_path)

    # note that the store is already coupled to HQ in the sandbox/devspace
    # config.
    try:
        store = next(s for s in existing_stores if s["url"] == local_store_url)
        client.store.set(store["id"])
        info(f"Using store with id '{store['id']}'")
    except StopIteration:
        error(
            "Local algorithm store not found. Please register its resources manually."
        )
        return ""

    # register also the other users in the local store
    users_in_store = client.store.user.list()["data"]
    all_users = client.user.list()["data"]
    for user in all_users:
        if user["keycloak_id"] not in [u["keycloak_id"] for u in users_in_store]:
            summary += f"Registering user {user['username']} in local store\n"
            client.store.user.register(username=user["username"], roles=[1])

    # Remove existing algorithm
    # This is broken, see issue: https://github.com/vantage6/vantage6/issues/1824
    algorithms = client.algorithm.list(name="session basic example")["data"]
    if len(algorithms) > 0:
        summary += f"Removing existing algorithm {algorithms[0]['name']}\n"
        client.algorithm.delete(id_=algorithms[0]["id"])

    # Download the algorithm store from the github repo. This ensures that the data is
    # always up to date.
    try:
        url = (
            "https://raw.githubusercontent.com/vantage6/v6-session-basics/"
            "main/algorithm_store.json"
        )
        response = requests.get(url, timeout=10)
        response.raise_for_status()
        algorithm_json = response.json()
        function_metadata = algorithm_json["functions"]
    except requests.exceptions.RequestException as e:
        # Fallback to local file if download fails
        print(f"Warning: Could not download algorithm store from GitHub: {e}")
        print("Not putting the algorithm in the store.")
        function_metadata = []

    summary += "Creating Session Basics algorithm\n"
    client.algorithm.create(
        name="Session Basics",
        description="A set of basic algorithms for a session management",
        image="harbor2.vantage6.ai/algorithms/session-basics:latest",
        vantage6_version="5.0.0",
        code_url="https://github.com/vantage6-ai/v6-session-basics",
        partitioning="horizontal",
        functions=function_metadata or [],
    )

    # TODO get this json data by downloading it from the github repo - that ensures
    # that the data is always up to date.
    # This is part of issue #2042
    summary += "Creating Network Diagnostics algorithm\n"
    client.algorithm.create(
        name="Network Diagnostics",
        description=(
            "Functions to diagnose network policies, that is, to what extent the "
            "algorithm tasks running on a vantage6 node could have access to the "
            "internal k8s network or to the outside world."
        ),
        image="ghcr.io/hcadavid/v6-sessions-k8s-diagnostics:latest",
        vantage6_version="5.0.0",
        code_url="https://github.com/hcadavid/v6-sessions-k8s-diagnostics",
        partitioning="horizontal",
        functions=[
            {
                "name": "network_status",
                "display_name": "Get the network diagnostics performed on a given node",
                "standalone": True,
                "description": "Get the network diagnostics performed on a given node",
                "ui_visualizations": [],
                "step_type": AlgorithmStepType.FEDERATED_COMPUTE.value,
                "arguments": [
                    {
                        "has_default_value": True,
                        "name": "sleep_time",
                        "display_name": "Delay ",
                        "description": (
                            "Delay in seconds before finishing the the diagnostics "
                            "task (useful to keep the POD running and access it to "
                            "perform further tests) "
                        ),
                        "type": "integer",
                        "default_value": "1",
                        "is_frontend_only": False,
                    }
                ],
                "databases": [],
            },
            {
                "name": "central_network_diagnostics",
                "display_name": (
                    "Get the network diagnostics performed on all the nodes"
                ),
                "standalone": True,
                "description": (
                    "Get the network diagnostics performed on all the nodes within a "
                    "collaboration"
                ),
                "ui_visualizations": [],
                "step_type": AlgorithmStepType.CENTRAL_COMPUTE.value,
                "arguments": [
                    {
                        "has_default_value": True,
                        "name": "sleep_time",
                        "display_name": "Delay ",
                        "description": (
                            "Delay in seconds before finishing the the diagnostics "
                            "task (useful to keep the PODs running on all the nodes "
                            "to access them and perform further inspections) "
                        ),
                        "type": "integer",
                        "default_value": "1",
                        "is_frontend_only": False,
                    }
                ],
                "databases": [],
            },
        ],
    )

    return summary



import itertools
from pathlib import Path
from shutil import rmtree

import questionary as q

from vantage6.common import (
    error,
    info,
)
from vantage6.common.context import AppContext
from vantage6.common.globals import InstanceType

from vantage6.cli.common.utils import check_running
from vantage6.cli.globals import InfraComponentName
from vantage6.cli.utils import remove_file


def execute_remove(
    ctx: AppContext,
    instance_type: InstanceType,
    infra_component: InfraComponentName,
    name: str,
    system_folders: bool,
    force: bool,
) -> None:
    """
    Function to remove a vantage6 service instance.

    Parameters
    ----------
    ctx : AppContext
        App context object
    instance_type : InstanceType
        Type of instance that is checked
    infra_component : InfraComponentName
        The infra component of the service
    name : str
        Name of the instance
    system_folders : bool
        Whether to use system folders or user folders
    force : bool
        Whether to ask for confirmation before removing or not
    """
    if check_running(ctx.helm_release_name, instance_type, name, system_folders):
        error(
            f"The {infra_component.value} {name} is still running! Please stop the "
            f"{infra_component.value} before deleting it."
        )
        exit(1)

    if not force:
        if not q.confirm(
            f"This {infra_component.value} will be deleted permanently including its "
            "configuration. Are you sure?",
            default=False,
        ).ask():
            info(f"The {infra_component.value} {name} will not be deleted")
            exit(0)

    # remove the config file
    remove_file(ctx.config_file, "configuration")

    # ensure log files are closed before removing
    log_dir = Path(ctx.log_file.parent)
    if log_dir.exists():
        info(f"Removing log directory: {log_dir}")
        for handler in itertools.chain(ctx.log.handlers, ctx.log.root.handlers):
            handler.close()
        # remove the whole folder with all the log files. This may also still contain
        # other files like (for HQ) RabbitMQ configuration etc
        rmtree(log_dir)



import os
from pathlib import Path

import click
import copier.errors
import questionary as q
from copier import run_update

from vantage6.common import info, warning


@click.command()
@click.option(
    "-d",
    "--dir",
    "directory",
    default=None,
    type=str,
    help="Directory to put the algorithm into",
)
@click.option(
    "--change-answers",
    is_flag=True,
    flag_value=True,
    help="Change answers to questions that were already answered",
    default=False,
)
def cli_algorithm_update(directory: str, change_answers: bool) -> dict:
    """Update an algorithm template

    When a new version of the algorithm template is released, you can update
    your existing algorithm template (which you created with
    `v6 algorithm create`) by running this command.
    """
    if not directory:
        default_dir = str(Path(os.getcwd()))
        try:
            directory = q.text("Algorithm directory:", default=default_dir).unsafe_ask()
        except KeyboardInterrupt:
            info("Aborted by user!")
            return

    info("Updating template...")
    try:
        run_update(
            directory, overwrite=True, skip_answered=not change_answers, unsafe=True
        )
        info("Template updated!")
    except copier.errors.UserMessageError as exc:
        warning(f"Update failed: {exc}")



import click
import requests

from vantage6.common import error, info
from vantage6.common.globals import (
    DEFAULT_API_PATH,
    HTTP_LOCALHOST,
    InstanceType,
    Ports,
)

from vantage6.cli import __version__
from vantage6.cli.common.utils import extract_name_and_is_sandbox
from vantage6.cli.common.version import get_and_select_ctx
from vantage6.cli.globals import DEFAULT_API_SERVICE_SYSTEM_FOLDERS


@click.command()
@click.option("-n", "--name", default=None, help="Configuration name")
@click.option("--system", "system_folders", flag_value=True)
@click.option(
    "--user",
    "system_folders",
    flag_value=False,
    default=DEFAULT_API_SERVICE_SYSTEM_FOLDERS,
)
@click.option("--context", default=None, help="Kubernetes context to use")
@click.option("--namespace", default=None, help="Kubernetes namespace to use")
@click.option(
    "--sandbox", "is_sandbox", flag_value=True, help="Is this a sandbox environment?"
)
def cli_algo_store_version(
    name: str, system_folders: bool, context: str, namespace: str, is_sandbox: bool
) -> None:
    """
    Print the version of the vantage6 algorithm store.
    """
    name, is_sandbox = extract_name_and_is_sandbox(name, is_sandbox)
    if is_sandbox:
        system_folders = False

    ctx = get_and_select_ctx(
        InstanceType.ALGORITHM_STORE,
        name,
        system_folders,
        context,
        namespace,
        is_sandbox,
    )
    store_config = ctx.config.get("store", {})

    if is_sandbox:
        port = Ports.SANDBOX_ALGO_STORE.value
    else:
        port = store_config.get("port", Ports.DEV_ALGO_STORE.value)
    api_path = store_config.get("api_path", DEFAULT_API_PATH)
    if not port:
        error("No port found in algorithm store configuration.")
        return
    if not api_path:
        error("No API path found in algorithm store configuration.")
        return

    response = requests.get(f"{HTTP_LOCALHOST}:{port}{api_path}/version")
    if response.status_code != 200:
        error("Failed to get algorithm store version.")
        return
    algorithm_store_version = response.json().get("version", "")

    info("")
    info(f"Algorithm store version: {algorithm_store_version}")
    info(f"CLI version: {__version__}")



import json
from typing import Any

from rich.console import Console
from rich.table import Table

from vantage6.common import debug, info
from vantage6.common.enum import AlgorithmStepType, TaskDatabaseType

from vantage6.client import UserClient

from vantage6.cli.globals import DIAGNOSTICS_IMAGE


class DiagnosticRunner:
    """
    Class to run the diagnostic algorithm on a vantage6 network.

    This class will create a task in the requested collaboration that will test
    the functionality of vantage6, and will report back the results.

    Parameters
    ----------
    client : UserClient
        The client to use for communication with the vantage6 hub.
    collaboration_id : int
        The ID of the collaboration to run the diagnostics in.
    organizations : list[int] | None
        The ID(s) of the organization(s) to run the diagnostics for. If None
        (default), run the diagnostics for all organizations in the
        collaboration.
    online_only : bool
        Whether to run the diagnostics only on nodes that are online. By
        default False
    session_id : int
        The ID of the session to use for the diagnostic test. By default 1.
    database_label : str
        The label of the database to use for the diagnostic test. By default "default".
    """

    def __init__(
        self,
        client: UserClient,
        collaboration_id: int,
        organizations: list[int] | None = None,
        online_only: bool = False,
        session_id: int = 1,
        database_label: str = "olympic-athletes",
    ) -> None:
        self.client = client
        self.collaboration_id = collaboration_id
        self.session_id = session_id
        self.database_label = database_label
        self.extract_task_details = None

        if not organizations:
            # run on all organizations in the collaboration
            # TODO use pagination properly, instead of just getting the first
            # 1000 organizations (which very likely is enough though)
            orgs = self.client.organization.list(
                collaboration=self.collaboration_id, per_page=1000
            )
            self.organization_ids = [org["id"] for org in orgs["data"]]
        else:
            self.organization_ids = organizations

        if online_only:
            nodes = self.client.node.list(
                collaboration=self.collaboration_id, is_online=True
            )
            debug(nodes)
            online_orgs = [node["organization"]["id"] for node in nodes["data"]]
            self.organization_ids = list(
                set(self.organization_ids).intersection(online_orgs)
            )

        info(f"Running diagnostics to {len(self.organization_ids)} organization(s)")
        info(f"  organizations: {self.organization_ids}")
        info(f"  collaboration: {self.collaboration_id}")

    def __call__(self) -> Any:
        """
        Run the diagnostics.

        Parameters
        ----------
        base: bool
            Whether to run the base features of the diagnostic algorithm. By
            default True.
        """
        # first create extraction task
        if not self.extract_task_details:
            self.extract()
        # then create base features task
        base_results = self.base_features()
        return base_results

    def base_features(self) -> list[dict]:
        """
        Create a task to run the base features of the diagnostic algorithm.

        Returns
        -------
        list[dict]
            The results of the diagnostic algorithm.
        """
        info("Starting task to test base features...")
        task = self.client.task.create(
            collaboration=self.collaboration_id,
            name="test",
            description="Basic Diagnostic test",
            image=DIAGNOSTICS_IMAGE,
            method="base_features",
            organizations=self.organization_ids,
            databases=[
                {
                    "dataframe_id": self.extraction_task_details.get("id"),
                    "type": TaskDatabaseType.DATAFRAME,
                }
            ],
            session=self.session_id,
            action=AlgorithmStepType.CENTRAL_COMPUTE,
        )
        debug(task)

        return self._wait_and_display(task.get("id"))

    def extract(self) -> None:
        """
        Create a task to extract the database.
        """
        info("Before running compute task, we need to create a dataframe")
        info(f"Creating dataframe for database with label: {self.database_label}")
        self.extraction_task_details = self.client.dataframe.create(
            label=self.database_label,
            image=DIAGNOSTICS_IMAGE,
            method="read_csv",
            arguments={},
            session=self.session_id,
        )

        self.client.wait_for_results(
            self.extraction_task_details.get("last_session_task", {}).get("id")
        )
        info("Dataframe created successfully!")

    def _wait_and_display(self, task_id: int) -> list[dict]:
        """
        Wait for the task to finish and then display the results.

        Parameters
        ----------
        task_id : int
            The ID of the task to wait for.

        Returns
        -------
        list[dict]
            The results of the diagnostic algorithm.
        """
        # TODO should we have the option to combine these in one request? Seems
        # like it would be more efficient
        # TODO ensure that we get all pages of results
        results = self.client.wait_for_results(task_id=task_id)["data"]
        runs = self.client.run.from_task(task_id=task_id)["data"]
        print("\n")
        for res in results:
            matched_run = [run for run in runs if run["id"] == res["run"]["id"]][0]
            self.display_diagnostic_results(res, matched_run["organization"]["id"])
            print()
        return results

    def display_diagnostic_results(self, result: dict, org_id: int) -> None:
        """
        Print the results of the diagnostic algorithm.

        Parameters
        ----------
        result : list[dict]
            The result of the diagnostic algorithm.
        org_id : int
            The ID of the organization that the diagnostic algorithm was run on
        """
        res = json.loads(result["result"])
        t_ = Table(title=f"Basic Diagnostics Summary (organization {org_id})")
        t_.add_column("name")
        t_.add_column("success")
        e_ = Table(title=f"Basic Diagnostics Errors (organization {org_id})")
        e_.add_column("name")
        e_.add_column("exception")
        e_.add_column("traceback")
        e_.add_column("payload")
        errors = False
        for diag in res:
            if diag["success"]:
                success = ":heavy_check_mark: [green]success[/green]"
            else:
                success = ":x: [red]failed[/red]"
                e_.add_row(
                    diag["name"], diag["exception"], diag["traceback"], diag["payload"]
                )
                errors = True
            t_.add_row(diag["name"], success)

        console = Console()
        console.print(t_)
        if errors:
            console.print(e_)



import subprocess
import time

from vantage6.common import error, info, warning

from vantage6.cli.k8s_config import KubernetesConfig


def run_kubectl_command(
    args: list[str],
    k8s_config: KubernetesConfig,
    check: bool = True,
    use_k8s_config_namespace: bool = True,
    input_text: str | None = None,
) -> subprocess.CompletedProcess:
    """
    Run a kubectl command with the appropriate context and namespace.

    Parameters
    ----------
    args : list[str]
        The kubectl arguments to run.
    k8s_config: KubernetesConfig
        Kubernetes configuration object with context and namespace.
    check : bool
        Whether to raise an exception on non-zero exit code.
    use_k8s_config_namespace: bool
        Whether to use the namespace from the Kubernetes configuration.
    input_text: str | None
        Optional stdin content to pass to kubectl (for example when using
        ``kubectl apply -f -``).

    Returns
    -------
    subprocess.CompletedProcess
        The result of the subprocess call.

    Raises
    ------
    subprocess.CalledProcessError
        If the command fails and check is True.
    """
    command = ["kubectl"] + args
    if k8s_config.context:
        command.extend(["--context", k8s_config.context])

    if k8s_config.namespace and use_k8s_config_namespace:
        command.extend(["--namespace", k8s_config.namespace])

    try:
        result = subprocess.run(
            command,
            check=check,
            capture_output=True,
            text=True,
            input=input_text,
        )
        return result
    except subprocess.CalledProcessError as e:
        error(f"Command failed: {' '.join(command)}")
        if e.stderr:
            error(f"Error: {e.stderr}")
        if check:
            raise
        return e


def wait_for_pod_ready(
    selector: str, description: str, k8s_config: KubernetesConfig
) -> None:
    """
    Wait for at least one pod matching the selector to be created and become Ready.

    Parameters
    ----------
    selector : str
        The selector to use to find the pods.
    description : str
        The description of the pods to wait for.
    k8s_config : KubernetesConfig
        The Kubernetes configuration to use.
    """

    info(f"Waiting for {description} pod(s) to be created...")
    attempt = 0
    max_attempts = 20
    while attempt < max_attempts:
        result = run_kubectl_command(
            ["get", "pod", "-l", selector, "-o", "name"],
            k8s_config,
            check=False,
        )
        if result.stdout.strip():
            break
        time.sleep(2)
        attempt += 1
        if attempt == max_attempts:
            error(f"Timeout while waiting for {description} pod(s) to be created.")
            exit(1)

    info(f"{description} pod(s) created, waiting for them to become Ready...")
    try:
        run_kubectl_command(
            [
                "wait",
                "--for=condition=ready",
                "pod",
                "-l",
                selector,
                "--timeout",
                "600s",
            ],
            k8s_config,
            check=True,
        )
    except subprocess.CalledProcessError:
        warning(f"Timeout while waiting for {description} pod(s) to become Ready.")



import os
from dataclasses import dataclass
from pathlib import Path
from typing import Callable

import appdirs
import questionary
import yaml
from colorama import Fore, Style
from kubernetes import config as k8s_config

from vantage6.common import error, info, warning
from vantage6.common.enum import StrEnumBase
from vantage6.common.globals import APPNAME

from vantage6.cli.globals import DEFAULT_CLI_CONFIG_FILE
from vantage6.cli.utils_kubernetes import get_core_api_with_ssl_handling

_CONTEXT_INFO_PRINTED = False


class K8SConfigVariable(StrEnumBase):
    """
    Enum containing the variables of the Kubernetes configuration.
    """

    CONTEXT = "context"
    NAMESPACE = "namespace"
    K8S_NODE = "k8s node"


@dataclass
class KubernetesConfig:
    """
    Class to store Kubernetes configuration.

    Attributes
    ----------
    context : str or None
        The last used Kubernetes context.
    namespace : str or None
        The last used Kubernetes namespace.
    k8s_node : str or None
        The last used Kubernetes node.
    """

    context: str | None = None
    namespace: str | None = None
    k8s_node: str | None = None

    def to_dict(self) -> dict:
        """
        Convert the Kubernetes configuration to a dictionary.
        """
        output = {}
        if self.context:
            output["context"] = self.context
        if self.namespace:
            output["namespace"] = self.namespace
        if self.k8s_node:
            output["k8s_node"] = self.k8s_node
        return output


class K8SConfigManager:
    """
    A class to manage CLI configuration for Kubernetes.

    The Kubernetes CLI configuration is stored in a YAML file, and contains, for
    example, the last used context and namespace.

    Attributes
    ----------
    config_path : PathLike
        Path to the configuration file.
    kubernetes_config : KubernetesConfig
        The Kubernetes configuration.
    _cached_mtime : float or None
        Last modification time of the configuration file.
    """

    def __init__(self) -> None:
        """
        Initialize the K8SConfigManager object.

        Parameters
        ----------
        config_path : str or PathLike, optional
            Path to the configuration file.
        """
        dirs = appdirs.AppDirs(APPNAME)
        self.config_path = Path(dirs.user_config_dir) / DEFAULT_CLI_CONFIG_FILE
        self._cached_mtime: float | None = None
        self.kubernetes_config = KubernetesConfig()

    def _load_config(self) -> None:
        """
        Load the configuration from the file.
        """
        loaded_config = {}
        if self.config_path.exists():
            with open(self.config_path, "r") as config_file:
                loaded_config = yaml.safe_load(config_file)

        context = loaded_config.get("context")
        namespace = loaded_config.get("namespace")
        k8s_node = loaded_config.get("k8s_node")
        self.kubernetes_config = KubernetesConfig(
            context=context,
            namespace=namespace,
            k8s_node=k8s_node,
        )

    def _save_config(self) -> None:
        """
        Save the configuration to the kubernetes configuration file.
        """
        self.config_path.parent.mkdir(parents=True, exist_ok=True)
        with open(self.config_path, "w") as config_file:
            yaml.dump(self.kubernetes_config.to_dict(), config_file)

    def _reload_cache_lazy(self) -> None:
        """
        Reload the configuration cache if the file has been modified.
        """
        if self.config_path.exists():
            mtime = os.path.getmtime(self.config_path)
            if self._cached_mtime != mtime:
                self._load_config()
                self._cached_mtime = mtime
        else:
            self.kubernetes_config = KubernetesConfig()
            self._cached_mtime = None

    def _set_context(self, context: str) -> None:
        """
        Set the Kubernetes context.

        Parameters
        ----------
        context : str
            The Kubernetes context to set.
        """
        if self.kubernetes_config.context != context:
            self.kubernetes_config.context = context
            self._save_config()
            self._reload_cache_lazy()

    def _set_namespace(self, namespace: str) -> None:
        """
        Set the Kubernetes namespace.

        Parameters
        ----------
        namespace : str
            The Kubernetes namespace to set.
        """
        if self.kubernetes_config.namespace != namespace:
            self.kubernetes_config.namespace = namespace
            self._save_config()
            self._reload_cache_lazy()

    def _set_k8s_node(self, k8s_node: str) -> None:
        """
        Set the Kubernetes node.
        """
        if self.kubernetes_config.k8s_node != k8s_node:
            self.kubernetes_config.k8s_node = k8s_node
            self._save_config()
            self._reload_cache_lazy()

    def get_active_settings(
        self, input_k8s_config: KubernetesConfig
    ) -> KubernetesConfig:
        """
        Get the active Kubernetes configuration.

        Parameters
        ----------
        input_k8s_config : KubernetesConfig
            The Kubernetes configuration to use.

        Returns
        -------
        KubernetesConfig
            An object containing the active context, namespace and k8s node.
        """
        # start by creating copy of the input_k8s_config
        active_k8s_config = KubernetesConfig(
            context=input_k8s_config.context,
            namespace=input_k8s_config.namespace,
            k8s_node=input_k8s_config.k8s_node,
        )

        _all_contexts, active_context = k8s_config.list_kube_config_contexts()
        if not active_k8s_config.context:
            active_k8s_config.context = active_context["name"]

        if not active_k8s_config.namespace:
            active_k8s_config.namespace = active_context["context"].get(
                "namespace", "default"
            )

        if not active_k8s_config.k8s_node:
            # Get node names from Kubernetes API
            try:
                core_api = get_core_api_with_ssl_handling()
                nodes = core_api.list_node()
                if nodes.items:
                    active_k8s_config.k8s_node = nodes.items[0].metadata.name
                else:
                    active_k8s_config.k8s_node = None
            except Exception:
                error("Failed to get Kubernetes nodes")
                active_k8s_config.k8s_node = None

        return active_k8s_config

    def _compare_config_variable(
        self,
        variable: K8SConfigVariable,
        current_value: str,
        last_value: str,
        set_func: Callable,
    ) -> None:
        """
        Compare a configuration variable with the last used value.

        Parameters
        ----------
        variable : K8SConfigVariables
            The variable to compare.
        current_value : str
            The current value of the variable.
        last_value : str
            The last used value of the variable.
        """
        if not last_value:
            set_func(current_value)
        elif last_value != current_value:
            warning(f"Are you using the correct {variable.value}?")
            warning(
                f"Current {variable.value}: {Fore.YELLOW}{current_value}"
                f"{Style.RESET_ALL}"
            )
            warning(
                f"Last    {variable.value}: {Fore.YELLOW}{last_value}{Style.RESET_ALL}"
            )

            new_value = questionary.select(
                f"Which {variable.value} do you want to use?",
                choices=[current_value, last_value],
                default=current_value,
            ).unsafe_ask()
            set_func(new_value)

    def select_k8s_config(
        self,
        input_k8s_config: KubernetesConfig,
    ) -> KubernetesConfig:
        """
        Compare active settings with last used settings.

        Parameters
        ----------
        input_k8s_config : KubernetesConfig
            The Kubernetes configuration to use.

        Returns
        -------
        KubernetesConfig
            An object containing the active context, namespace and k8s node.
        """

        active_k8s_config = self.get_active_settings(input_k8s_config)
        self._load_config()

        # compare context
        self._compare_config_variable(
            variable=K8SConfigVariable.CONTEXT,
            current_value=active_k8s_config.context,
            last_value=self.kubernetes_config.context,
            set_func=self._set_context,
        )

        # compare namespace
        self._compare_config_variable(
            variable=K8SConfigVariable.NAMESPACE,
            current_value=active_k8s_config.namespace,
            last_value=self.kubernetes_config.namespace,
            set_func=self._set_namespace,
        )

        # compare k8s node
        self._compare_config_variable(
            variable=K8SConfigVariable.K8S_NODE,
            current_value=active_k8s_config.k8s_node,
            last_value=self.kubernetes_config.k8s_node,
            set_func=self._set_k8s_node,
        )

        # only print the context and namespace once. This is to avoid printing it many
        # times, e.g. in sandbox commands
        global _CONTEXT_INFO_PRINTED
        if not _CONTEXT_INFO_PRINTED:
            info(
                f"Using    context: {Fore.YELLOW}{self.kubernetes_config.context}"
                f"{Style.RESET_ALL}"
            )
            info(
                f"Using  namespace: {Fore.YELLOW}"
                f"{self.kubernetes_config.namespace}{Style.RESET_ALL}"
            )
            info(
                f"Using   k8s node: {Fore.YELLOW}"
                f"{self.kubernetes_config.k8s_node}{Style.RESET_ALL}"
            )
            _CONTEXT_INFO_PRINTED = True

        return self.kubernetes_config


def get_k8s_node_names() -> list[str]:
    """
    Returns a list of available Kubernetes node names.

    Returns
    -------
    list[str]
        A list of available Kubernetes node names.
    """
    core_api = get_core_api_with_ssl_handling()
    nodes = core_api.list_node()

    return [item.metadata.name for item in nodes.items]


def select_k8s_config(
    context: str | None = None,
    namespace: str | None = None,
) -> KubernetesConfig:
    """
    Select the Kubernetes context and namespace.

    Parameters
    ----------
    context : str or None, optional
        The Kubernetes context to use.
    namespace : str or None, optional
        The Kubernetes namespace to use.

    Returns
    -------
    KubernetesConfig
        Object with the selected Kubernetes configuration.
    """
    k8s_config_manager = K8SConfigManager()
    return k8s_config_manager.select_k8s_config(
        KubernetesConfig(context=context, namespace=namespace)
    )



"""
Script to populate the hub with basic fixtures.
"""

import time
import traceback
from logging import info

import requests

from vantage6.common import error
from vantage6.common.globals import Ports

from vantage6.client import Client
from vantage6.client.utils import LogLevel

from vantage6.cli.sandbox.populate.helpers.connect_store import connect_store
from vantage6.cli.sandbox.populate.helpers.delete_fixtures import delete_fixtures
from vantage6.cli.sandbox.populate.helpers.load_fixtures import create_fixtures
from vantage6.cli.sandbox.populate.helpers.utils import NodeConfigCreationDetails


def populate_hub_dev(
    hq_url: str,
    auth_url: str,
    number_of_nodes: int,
    node_config_creation_details: NodeConfigCreationDetails,
) -> str | None:
    """
    Populate the hub with basic fixtures.

    Parameters
    ----------
    hq_url : str
        The URL of the HQ to connect to.
    auth_url : str
        The URL of the auth service to connect to.
    number_of_nodes : int
        The number of nodes to create.
    node_config_creation_details : NodeConfigCreationDetails | None
        The details of the node config creation. If None, the node configs will not be
        created.
    clear_dev_folders : bool
        Whether to clear everything inthe dev directory before creating the new
        resources.
    create_node_configs: bool
        Whether to create the node configs. Default is True.

    Returns
    -------
    str | None
        A report of the creation process, or None if an error occurred.

    Raises
    ------
    Exception
        If connection/authentication to the HQ fails.
    """
    client = _initalize_client(hq_url, auth_url)

    # Create new resources in the HQ and store
    # Delete existing resources in the HQ first, before creating new ones.
    try:
        report_deletion = delete_fixtures(client)
        report_creation = create_fixtures(
            client,
            number_of_nodes,
            node_config_creation_details=node_config_creation_details,
            clear_dev_folders=True,
        )
        report_store = connect_store(client)

        return report_deletion + "\n" + report_creation + "\n" + report_store

    except Exception:
        error("=" * 80)
        error("Failed to populate hub")
        error(traceback.format_exc())
        error("=" * 80)


def populate_hub_sandbox(
    hq_url: str,
    auth_url: str,
    number_of_nodes: int,
) -> dict:
    """
    Populate sandbox hub with basic resources.

    Parameters
    ----------
    hq_url : str
        The URL of the HQ to connect to.
    auth_url : str
        The URL of the auth service to connect to.
    number_of_nodes : int
        The number of nodes to create.

    Returns
    -------
    dict
        A dictionary containing the report of the creation process.

    Raises
    ------
    Exception
        If connection/authentication to the HQ fails.
    """
    client = _initalize_client(hq_url, auth_url)

    try:
        delete_fixtures(client)
        report_creation = create_fixtures(
            client,
            number_of_nodes,
            clear_dev_folders=False,
            return_as_dict=True,
        )
        connect_store(client, store_port=Ports.SANDBOX_ALGO_STORE.value)
    except Exception:
        error("=" * 80)
        error("Failed to populate hub")
        error(traceback.format_exc())
        error("=" * 80)
        exit(1)

    # return the details of the created nodes so that config files can be created
    return report_creation["nodes"]["created"]


def _wait_for_keycloak_realm(
    auth_url: str, realm_name: str = "vantage6", timeout: int = 120
) -> None:
    """
    Wait for Keycloak realm to be available before attempting authentication.

    This prevents race conditions where authentication is attempted before
    the realm import job has completed.

    Parameters
    ----------
    auth_url : str
        The URL of the Keycloak auth service.
    realm_name : str
        The name of the realm to wait for. Default is "vantage6".
    timeout : int
        Maximum time to wait in seconds. Default is 120.
    """
    info("Waiting for Keycloak realm to be ready...")
    well_known_url = f"{auth_url}/realms/{realm_name}/.well-known/openid-configuration"
    start_time = time.time()
    attempt = 0

    while (time.time() - start_time) < timeout:
        attempt += 1
        try:
            resp = requests.get(well_known_url, timeout=5)
            if resp.status_code == 200:
                info("Keycloak realm is ready.")
                return
            elif resp.status_code == 404:
                # Realm doesn't exist yet - wait and retry
                if attempt % 5 == 0:  # Log every 5th attempt
                    info(f"Realm not ready yet (attempt {attempt})...")
                time.sleep(2)
            else:
                # Other HTTP errors - log and retry
                if attempt % 10 == 0:  # Log every 10th attempt
                    info(f"Keycloak returned status {resp.status_code}, retrying...")
                time.sleep(2)
        except requests.exceptions.RequestException:
            # Network errors - wait and retry
            if attempt % 5 == 0:
                info(f"Keycloak not reachable yet (attempt {attempt})...")
            time.sleep(2)

    error(
        f"Timeout: Keycloak realm '{realm_name}' did not become available "
        f"within {timeout} seconds. The realm import job may have failed."
    )
    raise TimeoutError(
        f"Keycloak realm '{realm_name}' not ready after {timeout} seconds"
    )


def _initalize_client(hq_url: str, auth_url: str) -> Client:
    """
    Initialize an authenticated client to the HQ.

    The HQ may not be ready yet, so we retry until it is.

    Parameters
    ----------
    hq_url : str
        The URL of the HQ to connect to.
    auth_url : str
        The URL of the auth service to connect to.

    Returns
    -------
    Client
        An authenticated client to the vantage6 hub.
    """
    # Wait for Keycloak realm to be ready before attempting authentication
    _wait_for_keycloak_realm(auth_url)

    client = Client(
        hq_url=hq_url,
        auth_url=auth_url,
        log_level=LogLevel.WARN,
    )

    info("Waiting for authentication...")
    max_attempts = 120
    attempt = 1

    while attempt <= max_attempts:
        try:
            print(".", end="", flush=True)
            client.authenticate()
            info("Successfully authenticated!")
            break
        except Exception as e:
            if attempt == max_attempts:
                error(
                    f"Failed to authenticate after {max_attempts} attempts. "
                    "Vantage6 hub may not be online."
                )
                raise e

            time.sleep(5)
            attempt += 1

    return client



from pathlib import Path
from shutil import rmtree

import click

from vantage6.common.globals import InstanceType

from vantage6.cli.common.decorator import click_insert_context
from vantage6.cli.common.remove import execute_remove
from vantage6.cli.context.node import NodeContext
from vantage6.cli.globals import InfraComponentName


@click.command()
@click.option(
    "-f", "--force", type=bool, flag_value=True, help="Don't ask for confirmation"
)
@click.option("--sandbox/--no-sandbox", "sandbox", default=False)
@click_insert_context(
    type_=InstanceType.NODE,
    include_name=True,
    include_system_folders=True,
    sandbox_param="sandbox",
)
def cli_node_remove(
    ctx: NodeContext, name: str, system_folders: bool, force: bool
) -> None:
    """
    Delete a node permanently.

    Remove the configuration file, log file, and kubernetes resources attached to
    the node.
    """

    execute_remove(
        ctx, InstanceType.NODE, InfraComponentName.NODE, name, system_folders, force
    )

    # remove the folder: if it hasn't been started yet this won't exist...
    if Path.exists(ctx.config_dir / name):
        rmtree(ctx.config_dir / name)



import click
import requests

from vantage6.common import error, info
from vantage6.common.globals import InstanceType

from vantage6.cli import __version__
from vantage6.cli.common.utils import extract_name_and_is_sandbox
from vantage6.cli.common.version import get_and_select_ctx
from vantage6.cli.globals import DEFAULT_API_SERVICE_SYSTEM_FOLDERS


@click.command()
@click.option("-n", "--name", default=None, help="Configuration name")
@click.option("--system", "system_folders", flag_value=True)
@click.option(
    "--user",
    "system_folders",
    flag_value=False,
    default=DEFAULT_API_SERVICE_SYSTEM_FOLDERS,
)
@click.option("--context", default=None, help="Kubernetes context to use")
@click.option("--namespace", default=None, help="Kubernetes namespace to use")
@click.option(
    "--sandbox", "is_sandbox", flag_value=True, help="Is this a sandbox environment?"
)
def cli_hq_version(
    name: str, system_folders: bool, context: str, namespace: str, is_sandbox: bool
) -> None:
    """
    Print the version of the vantage6 HQ.
    """
    name, is_sandbox = extract_name_and_is_sandbox(name, is_sandbox)
    if is_sandbox:
        system_folders = False
    ctx = get_and_select_ctx(
        InstanceType.HQ, name, system_folders, context, namespace, is_sandbox
    )
    hq_config = ctx.config.get("hq", {})
    base_url = hq_config.get("baseUrl", "")
    api_path = hq_config.get("apiPath", "")
    if not base_url:
        error("No base URL found in HQ configuration.")
        return
    if not api_path:
        error("No API path found in HQ configuration.")
        return

    response = requests.get(f"{base_url}{api_path}/version")
    if response.status_code != 200:
        error("Failed to get HQ version.")
        return
    hq_version = response.json().get("version", "")

    info("")
    info(f"HQ version: {hq_version}")
    info(f"CLI version: {__version__}")



# from click import UsageError
# from vantage6.cli.test.client_script import cli_test_client_script

# import click
# import unittest


# class TestScriptTest(unittest.TestCase):
#     def test_script_incorrect_usage(self):
#         ctx = click.Context(cli_test_client_script)

#         with self.assertRaises(UsageError):
#             ctx.invoke(
#                 cli_test_client_script,
#                 script="path/to/script.py",
#                 task_arguments="{'my_arg': 1}",
#             )

#         with self.assertRaises(UsageError):
#             ctx.invoke(
#                 cli_test_client_script,
#                 task_arguments="not_a_json",
#             )



import logging

from kubernetes import client as k8s_client
from kubernetes.client.rest import ApiException

from vantage6.common import logger_name

log = logging.getLogger(logger_name(__name__))


def delete_run_related_pods(
    run_id: int,
    container_name: str,
    namespace: str,
    core_api: k8s_client.CoreV1Api,
    batch_api: k8s_client.BatchV1Api,
) -> None:
    """
    Deletes all the PODs created by a Kubernetes job in a given namespace

    Parameters
    ----------
    run_id: int
        ID of the run
    container_name: str
        Name of the container
    namespace: str
        Namespace where the container is located
    core_api: k8s_client.CoreV1Api
        Kubernetes Core API instance
    batch_api: k8s_client.BatchV1Api
        Kubernetes Batch API instance
    """
    log.info(
        "Cleaning up kubernetes Job %s (run_id = %s) and related PODs",
        container_name,
        run_id,
    )

    __delete_job(container_name, namespace, batch_api)

    job_selector = f"job-name={container_name}"
    job_pods_list = core_api.list_namespaced_pod(namespace, label_selector=job_selector)
    for job_pod in job_pods_list.items:
        __delete_pod(job_pod.metadata.name, namespace, core_api)

    __delete_secret(container_name, namespace, core_api)
    # delete secret with Docker login credentials for private Docker registries
    # if it exists
    __delete_secret(
        f"docker-login-secret-run-id-{run_id}",
        namespace,
        core_api,
        log_if_not_found=False,
    )


def __delete_secret(
    secret_name: str,
    namespace: str,
    core_api: k8s_client.CoreV1Api,
    log_if_not_found: bool = True,
) -> None:
    """
    Deletes a secret in a given namespace

    Parameters
    ----------
    secret_name: str
        Name of the secret
    namespace: str
        Namespace where the secret is located
    core_api: k8s_client.CoreV1Api
        Kubernetes Core API instance
    log_if_not_found: bool
        If True, it will be logged if the secret does not exist. By default, True.
    """
    try:
        core_api.delete_namespaced_secret(name=secret_name, namespace=namespace)
        log.info(
            "Removed kubernetes Secret %s in namespace %s",
            secret_name,
            namespace,
        )
    except ApiException as exc:
        if exc.status == 404:
            if log_if_not_found:
                log.debug(
                    "No secret %s to remove in namespace %s", secret_name, namespace
                )
        else:
            log.error("Exception when deleting namespaced secret: %s", exc)


def __delete_job(
    job_name: str, namespace: str, batch_api: k8s_client.BatchV1Api
) -> None:
    """
    Deletes a job in a given namespace

    Parameters
    ----------
    job_name: str
        Name of the job
    namespace: str
        Namespace where the job is located
    batch_api: k8s_client.BatchV1Api
        Kubernetes Batch API instance
    """
    log.info(
        "Cleaning up kubernetes Job %s and related PODs",
        job_name,
    )
    try:
        # Check if the job exists before attempting to delete it
        job = batch_api.read_namespaced_job(name=job_name, namespace=namespace)
        if job:
            batch_api.delete_namespaced_job(name=job_name, namespace=namespace)
        else:
            log.warning(
                "Job %s not found in namespace %s, skipping deletion",
                job_name,
                namespace,
            )
    except ApiException as exc:
        if exc.status == 404:
            log.warning(
                "Job %s not found in namespace %s, skipping deletion",
                job_name,
                namespace,
            )
        else:
            log.error("Exception when deleting namespaced job: %s", exc)


def __delete_pod(pod_name: str, namespace: str, core_api: k8s_client.CoreV1Api) -> None:
    """
    Deletes a job in a given namespace

    Parameters
    ----------
    pod_name: str
        Name of the job
    namespace: str
        Namespace where the job is located
    core_api: k8s_client.CoreV1Api
        Kubernetes Core API instance
    """
    log.info("Cleaning up kubernetes pod %s in namespace %s", pod_name, namespace)
    try:
        # Check if the job exists before attempting to delete it
        job = core_api.read_namespaced_pod(name=pod_name, namespace=namespace)
        if job:
            core_api.delete_namespaced_pod(name=pod_name, namespace=namespace)
        else:
            log.warning(
                "Pod %s not found in namespace %s, skipping deletion",
                pod_name,
                namespace,
            )
    except ApiException as exc:
        if exc.status == 404:
            log.warning(
                "Pod %s not found in namespace %s, skipping deletion",
                pod_name,
                namespace,
            )
        else:
            log.error("Exception when deleting namespaced job: %s", exc)



import click

from vantage6.common.globals import InstanceType

from vantage6.cli.common.decorator import click_insert_context
from vantage6.cli.common.remove import execute_remove
from vantage6.cli.context import AlgorithmStoreContext
from vantage6.cli.globals import InfraComponentName


@click.command()
@click.option("--sandbox/--no-sandbox", "sandbox", default=False)
@click_insert_context(
    type_=InstanceType.ALGORITHM_STORE,
    include_name=True,
    include_system_folders=True,
    sandbox_param="sandbox",
)
@click.option("-f", "--force", "force", flag_value=True)
def cli_algo_store_remove(
    ctx: AlgorithmStoreContext, name: str, system_folders: bool, force: bool
) -> None:
    """
    Function to remove an algorithm store.

    Parameters
    ----------
    ctx : AlgorithmStoreContext
        Algorithm store context object
    force : bool
        Whether to ask for confirmation before removing or not
    """
    execute_remove(
        ctx,
        InstanceType.ALGORITHM_STORE,
        InfraComponentName.ALGORITHM_STORE,
        name,
        system_folders,
        force,
    )



from __future__ import annotations

import subprocess
from collections.abc import Callable

from colorama import Fore, Style

from vantage6.common import error, info
from vantage6.common.globals import InstanceType

from vantage6.cli.common.utils import (
    extract_name_and_is_sandbox,
    find_running_service_names,
    select_running_service,
)
from vantage6.cli.context import get_context
from vantage6.cli.globals import CLICommandName, InfraComponentName
from vantage6.cli.k8s_config import KubernetesConfig, select_k8s_config
from vantage6.cli.utils import validate_input_cmd_args


def execute_cli_stop(
    command_name: CLICommandName,
    name: str,
    k8s_config: KubernetesConfig,
    system_folders: bool,
    is_sandbox: bool = False,
) -> None:
    """
    Execute the stop command for an infrastructure service.
    """
    cmd = [
        "v6",
        command_name.value,
        "stop",
        "--name",
        name,
        "--context",
        k8s_config.context,
        "--namespace",
        k8s_config.namespace,
    ]
    cmd.append("--system" if system_folders else "--user")
    if is_sandbox:
        cmd.append("--sandbox")
    subprocess.run(cmd, check=True)


def execute_stop(
    stop_function: Callable,
    instance_type: InstanceType,
    infra_component: InfraComponentName,
    stop_all: bool,
    to_stop: str | None,
    namespace: str,
    context: str,
    system_folders: bool,
    is_sandbox: bool = False,
    stop_function_args: dict | None = None,
):
    """
    Execute the stop function for a given instance type and infra component.

    Parameters
    ----------
    stop_function : Callable
        The function to stop the service.
    instance_type : InstanceType
        The instance type of the service.
    infra_component : InfraComponentName
        The infra component of the service.
    stop_all : bool
        Whether to stop all services.
    to_stop : str | None
        The name of the service to stop. If None, the user will be asked to select a
        service.
    namespace : str
        The namespace of the service.
    context : str
        The context of the service.
    system_folders : bool
        Whether to use system folders.
    is_sandbox : bool
        Whether the configuration is a sandbox configuration, by default False
    stop_function_args : dict | None
        Additional arguments to pass to the stop function
    """
    if stop_function_args is None:
        stop_function_args = {}
    k8s_config = select_k8s_config(
        context=context,
        namespace=namespace,
    )
    running_services = find_running_service_names(
        instance_type=instance_type,
        only_system_folders=system_folders,
        only_user_folders=not system_folders,
        k8s_config=k8s_config,
    )

    if not running_services:
        error(f"No running {infra_component.value}s found.")
        return

    if stop_all:
        for service in running_services:
            stop_function(service, k8s_config.namespace, k8s_config.context)
    else:
        if not to_stop:
            helm_name = select_running_service(running_services, instance_type)
        else:
            to_stop, is_sandbox = extract_name_and_is_sandbox(to_stop, is_sandbox)
            ctx = get_context(
                instance_type, to_stop, system_folders, is_sandbox=is_sandbox
            )
            helm_name = ctx.helm_release_name

        if helm_name in running_services:
            stop_function(
                helm_name,
                k8s_config,
                **stop_function_args,
            )
            info(
                f"Stopped the {Fore.GREEN}{helm_name}{Style.RESET_ALL} "
                f"{infra_component.value}."
            )
        else:
            error(f"{Fore.RED}{to_stop}{Style.RESET_ALL} is not running?!")


def helm_uninstall(
    release_name: str, k8s_config: KubernetesConfig | None = None
) -> None:
    """
    Manage the `helm uninstall` command.

    Parameters
    ----------
    release_name : str
        The name of the Helm release to uninstall.
    k8s_config : KubernetesConfig, optional
        The Kubernetes configuration to use.
    """
    # Input validation
    validate_input_cmd_args(release_name, "release name")
    if k8s_config:
        validate_input_cmd_args(k8s_config.context, "context name", allow_none=True)
        validate_input_cmd_args(k8s_config.namespace, "namespace name", allow_none=True)

    # Create the command
    max_time_stop = "60s"
    info(
        f"Stopping helm chart '{release_name}'. This may take up to {max_time_stop}..."
    )
    command = ["helm", "uninstall", release_name, "--wait", "--timeout", max_time_stop]

    if k8s_config and k8s_config.context:
        command.extend(["--kube-context", k8s_config.context])

    if k8s_config and k8s_config.namespace:
        command.extend(["--namespace", k8s_config.namespace])

    try:
        subprocess.run(
            command,
            stdout=subprocess.DEVNULL,
            check=True,
        )
        info(f"Successfully uninstalled release '{release_name}'.")
    except subprocess.CalledProcessError as e:
        error(f"Failed to uninstall release '{release_name}': {e}")
    except FileNotFoundError:
        error(
            "Helm command not found. Please ensure Helm is installed and available in "
            "the PATH."
        )



import subprocess
import sys

import click

from vantage6.common import error, info

from vantage6.cli.dev.common import check_devspace_installed


@click.command()
def cli_stop_dev_env():
    """Stop the development environment."""
    check_devspace_installed()

    try:
        info("🛑 Stopping development environment with devspace...")
        cmd = ["devspace", "run", "stop-dev"]
        subprocess.run(cmd, check=True, capture_output=False)
        info("✅ Development environment stopped successfully!")
    except subprocess.CalledProcessError as e:
        error(f"❌ Error stopping development environment: {e}")
        sys.exit(e.returncode)



"""
Development script to delete all entities from HQ
"""

from vantage6.client import Client


def delete_fixtures(client: Client) -> str:
    """
    Delete all entities from HQ.

    Arguments
    ---------
    client: Client
        The client to use to delete the entities.
    """

    # Track deletion counts
    deletion_counts = {}

    # Delete sessions & dataframes
    sessions = client.session.list(per_page=999)["data"]
    for session in sessions:
        client.session.delete(session["id"], delete_dependents=True)
    deletion_counts["sessions"] = len(sessions)

    # Delete tasks and nodes
    for client_subclass_name in ("task", "node"):
        client_subclass = getattr(client, client_subclass_name)
        entities = client_subclass.list(per_page=999)["data"]
        for entity in entities:
            client_subclass.delete(entity["id"])
        deletion_counts[f"{client_subclass_name}s"] = len(entities)

    # Delete collaborations
    collabs = client.collaboration.list(per_page=999, scope="global")["data"]
    for collab in collabs:
        client.collaboration.delete(collab["id"])
    deletion_counts["collaborations"] = len(collabs)

    # Delete users (excluding admin)
    users = client.user.list(per_page=999)["data"]
    deleted_users = 0
    for user in users:
        if user["username"] == "admin":
            continue
        client.user.delete(user["id"])
        deleted_users += 1
    deletion_counts["users"] = deleted_users

    # Delete organizations (excluding the default organization with id 1)
    orgs = client.organization.list(per_page=999)["data"]
    deleted_orgs = 0
    for org in orgs:
        if org["id"] == 1:
            continue
        client.organization.delete(org["id"])
        deleted_orgs += 1
    deletion_counts["organizations"] = deleted_orgs

    # Create summary string
    summary = "=== Deletion Summary ===\n"
    for entity_type, count in deletion_counts.items():
        summary += f"Deleted {count} {entity_type}\n"
    summary += "======================="
    print(summary)
    return summary



import questionary as q

from vantage6.common.context import AppContext
from vantage6.common.globals import (
    InstanceType,
)

from vantage6.cli.configuration_create import add_common_backend_config


def algo_store_configuration_questionaire(
    instance_name: str, system_folders: bool
) -> dict:
    """
    Questionary to generate a config file for the algorithm store
    instance.

    Parameters
    ----------
    instance_name : str
        Name of the store instance.
    system_folders : bool
        Whether to use system folders or user folders.

    Returns
    -------
    dict
        Dictionary with the new store configuration
    """
    config = {"store": {"keycloak": {}}, "database": {}}

    config = add_common_backend_config(
        config, InstanceType.ALGORITHM_STORE, instance_name
    )

    # ask about openness of the algorithm store
    config["store"]["policies"] = {}
    is_open = q.confirm(
        "Do you want to open the algorithm store to the public? This will allow anyone "
        "to view the algorithms, but they cannot modify them.",
        default=False,
    ).unsafe_ask()
    if is_open:
        open_algos_policy = "public"
    else:
        is_open_to_whitelist = q.confirm(
            "Do you want to allow all authenticated users to access "
            "the algorithms in the store? If not allowing this, you will have to assign"
            " the appropriate permissions to each user individually.",
            default=True,
        ).unsafe_ask()
        open_algos_policy = "authenticated" if is_open_to_whitelist else "private"
    config["store"]["policies"]["algorithm_view"] = open_algos_policy

    dirs = AppContext.instance_folders(
        InstanceType.ALGORITHM_STORE, instance_name, system_folders
    )
    log_dir = str(dirs["log"])
    config["store"]["logging"]["volumeHostPath"] = log_dir

    return config



import click
import requests
import yaml
from marshmallow import Schema, ValidationError, fields, post_load

from vantage6.common import error, info
from vantage6.common.globals import (
    InstanceType,
)

from vantage6.client import UserClient
from vantage6.client.utils import LogLevel

from vantage6.cli import __version__
from vantage6.cli.common.decorator import click_insert_context
from vantage6.cli.context.hq import HQContext


@click.command()
@click.argument("file", type=click.Path(exists=True))
@click.option(
    "--drop-all",
    is_flag=True,
    default=False,
    help="Drop all existing data before importing",
)
@click.option("--sandbox/--no-sandbox", "sandbox", default=False)
@click_insert_context(type_=InstanceType.HQ, sandbox_param="sandbox")
def cli_hq_import(ctx: HQContext, file: str, drop_all: bool) -> None:
    """
    Import vantage6 resources into a HQ instance from a yaml FILE.

    This allows you to create organizations, collaborations, users, tasks, etc. from a
    YAML file. This method expects the YAML file to be located on the
    same machine as this method is invoked from.

    This import assigns the root role to all users, which contains all permissions. So
    use this with caution.

    The FILE argument should be a path to a yaml file containing the vantage6 formatted
    data to import.
    """
    info("Validating HQ version: ")
    info(f"{ctx.config['hq']['baseUrl']}{ctx.config['hq']['apiPath']}/version")
    response = requests.get(
        f"{ctx.config['hq']['baseUrl']}{ctx.config['hq']['apiPath']}/version"
    )
    if response.status_code != 200:
        error("Unable to get HQ version")
        return

    body = response.json()
    if "version" not in body:
        error("HQ gave a valid response but did not include the version")
        return

    # Compare it to this package version
    hq_version = body["version"]
    if hq_version != __version__:
        error(
            f"You are using CLI version {__version__} but the HQ is running "
            f"version {hq_version}. Please use the same version of the CLI and HQ."
        )
        return

    info("Loading and validating import file")
    with open(file, "r") as f:
        import_data = yaml.safe_load(f)

    _check_import_file(import_data)

    client = UserClient(
        hq_url=f"{ctx.config['hq']['baseUrl']}{ctx.config['hq']['apiPath']}",
        auth_url=ctx.config["ui"]["keycloak"]["publicUrl"],
        auth_realm=ctx.config["ui"]["keycloak"]["realm"],
        auth_client=ctx.config["ui"]["keycloak"]["client"],
        log_level=LogLevel.INFO,
    )

    info("Authenticate using admin credentials (opens browser for login)")
    client.authenticate()

    # Note: we do not validate that the user has the correct permissions to import data.
    # As the user has access to the `v6 hq import` command, they already have
    # access to HQ+database.

    if drop_all:
        info("Dropping all existing data")
        _drop_all(client)

    info("Collecting root role and rule")
    root_role_id = client.role.list(name="Root", include_root=True)["data"][0]["id"]

    info("Importing organizations")
    organizations = []
    for organization in import_data["organizations"]:
        org = client.organization.create(
            name=organization["name"],
            address1=organization.get("address1", ""),
            address2=organization.get("address2", ""),
            zipcode=str(organization.get("zipcode", "")),
            country=organization.get("country", ""),
            domain=organization.get("domain", ""),
            public_key=organization.get("public_key", ""),
        )
        organizations.append(org)

        info(f"Importing users for organization {org['name']}")
        for user in organization.get("users", []):
            client.user.create(
                username=user["username"],
                password=user["password"],
                organization=org["id"],
                roles=[root_role_id],
            )

    info("Importing collaborations")
    all_nodes = []
    for collaboration in import_data["collaborations"]:
        # Collecting organization ids
        organization_ids = []
        for participant in collaboration["participants"]:
            for org in organizations:
                if org["name"] == participant["name"]:
                    organization_ids.append(org["id"])

        col = client.collaboration.create(
            name=collaboration["name"],
            organizations=organization_ids,
            encrypted=collaboration.get("encrypted", False),
        )

        info("Registering nodes for collaboration")
        for participant in collaboration["participants"]:
            for org in organizations:
                if org["name"] == participant["name"]:
                    node = client.node.create(
                        name=f"{collaboration['name']}-{org['name'].replace(' ', '-')}-node",
                        organization=org["id"],
                        collaboration=col["id"],
                    )
                    all_nodes.append(node)

    return all_nodes


def _drop_all(client: UserClient) -> None:
    """
    Drop all existing data from the HQ.
    """
    while nodes := client.node.list()["data"]:
        for node in nodes:
            info(f"Deleting node {node['name']}")
            client.node.delete(node["id"])

    while collaborations := client.collaboration.list(scope="global")["data"]:
        for collaboration in collaborations:
            info(f"Deleting collaboration {collaboration['name']}")
            client.collaboration.delete(collaboration["id"], delete_dependents=True)

    # remove all organizations but keep the root organization
    while orgs := [o for o in client.organization.list()["data"] if o["id"] != 1]:
        for org in orgs:
            info(f"Deleting organization {org['name']}")
            client.organization.delete(org["id"], delete_dependents=True)


class StrOrInt(fields.Field):
    """Field that can be a string or an integer"""

    def _serialize(self, value, attr, obj, **kwargs):
        if isinstance(value, str):
            return value
        elif isinstance(value, int):
            return str(value)
        else:
            raise ValidationError("Value must be a string or an integer")


class UserSchema(Schema):
    username = fields.Str(required=True)
    password = fields.Str(required=True)


class OrganizationSchema(Schema):
    name = fields.Str(required=True)
    address1 = fields.Str(missing="")
    address2 = fields.Str(missing="")
    zipcode = StrOrInt(missing="")
    country = fields.Str(missing="")
    domain = fields.Str(missing="")
    public_key = fields.Str(missing="")
    users = fields.List(fields.Nested(UserSchema), missing=[])


class ParticipantSchema(Schema):
    name = fields.Str(required=True)


class CollaborationSchema(Schema):
    name = fields.Str(required=True)
    participants = fields.List(fields.Nested(ParticipantSchema), required=True)
    encrypted = fields.Bool(missing=False)


class ImportDataSchema(Schema):
    organizations = fields.List(fields.Nested(OrganizationSchema), missing=[])
    collaborations = fields.List(fields.Nested(CollaborationSchema), missing=[])

    @post_load
    def validate_participants_exist(self, data, **kwargs):
        """Validate that all participants exist in organizations"""
        org_names = {org["name"] for org in data.get("organizations", [])}

        for collaboration in data.get("collaborations", []):
            for participant in collaboration.get("participants", []):
                if participant["name"] not in org_names:
                    raise ValidationError(
                        f"Participant {participant['name']} not found in organizations",
                        field_name="collaborations",
                    )
        return data


def _check_import_file(import_data: dict) -> dict:
    """
    Validate import file using Marshmallow schemas.
    Returns the validated and cleaned data.
    """
    schema = ImportDataSchema()
    try:
        return schema.load(import_data)
    except ValidationError as err:
        # Handle validation errors gracefully
        error(f"Validation error: {err.messages}")
        exit(1)



import os
import unittest
from pathlib import Path
from unittest.mock import MagicMock, patch

from click.testing import CliRunner

from vantage6.common.globals import LOCALHOST, InstanceType, Ports

from vantage6.cli.globals import APPNAME, InfraComponentName
from vantage6.cli.k8s_config import KubernetesConfig
from vantage6.cli.node.attach import cli_node_attach
from vantage6.cli.node.common import create_client_and_authenticate
from vantage6.cli.node.create_private_key import cli_node_create_private_key
from vantage6.cli.node.files import cli_node_files
from vantage6.cli.node.list import cli_node_list
from vantage6.cli.node.new import cli_node_new_configuration
from vantage6.cli.node.restart import cli_node_restart
from vantage6.cli.node.start import cli_node_start
from vantage6.cli.node.stop import cli_node_stop


class NodeCLITest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        return super().setUpClass()

    @patch("vantage6.cli.context.node.NodeContext.available_configurations")
    @patch("vantage6.cli.common.list.find_running_service_names")
    def test_list(self, find_running_service_names, available_configurations):
        """A container list and their current status."""
        # https://docs.python.org/3/library/unittest.mock.html#mock-names-and-the-name-attribute

        node_name = "iknl"
        find_running_service_names.return_value = [f"{APPNAME}-{node_name}-user-node"]

        # returns a list of configurations and failed inports
        def side_effect(system_folders):
            config = MagicMock()
            config.name = node_name
            if not system_folders:
                return [[config], []]
            else:
                return [[config], []]

        available_configurations.side_effect = side_effect

        # invoke CLI method
        runner = CliRunner()
        result = runner.invoke(cli_node_list, [])

        # validate exit code
        self.assertEqual(result.exit_code, 0)

        # check printed lines
        self.assertEqual(
            result.output,
            "\nName                     Status          System/User\n"
            "-----------------------------------------------------\n"
            "iknl                     Not running     System \n"
            "iknl                     Running         User   \n"
            "-----------------------------------------------------\n",
        )

    @patch("vantage6.cli.common.new.make_configuration")
    @patch("vantage6.cli.common.new.ensure_config_dir_writable")
    @patch("vantage6.cli.node.common.NodeContext")
    def test_new_config(self, context, permissions, make_configuration):
        """No error produced when creating new configuration."""
        context.config_exists.return_value = False
        permissions.return_value = True
        make_configuration.return_value = (MagicMock(), "/some/file/path")

        runner = CliRunner()
        result = runner.invoke(
            cli_node_new_configuration,
            [
                "--name",
                "some-name",
            ],
        )

        # check that info message is produced
        self.assertEqual(result.output[:7], "[info ]")

        # check OK exit code
        self.assertEqual(result.exit_code, 0)

    @patch("vantage6.cli.common.new.make_configuration")
    def test_new_config_replace_whitespace_in_name(self, make_configuration):
        """Whitespaces are replaced in the name."""

        runner = CliRunner()
        result = runner.invoke(
            cli_node_new_configuration,
            [
                "--name",
                "some name",
            ],
        )

        self.assertTrue("[error] - Invalid name: 'some name'." in result.output)
        self.assertEqual(result.exit_code, 1)

    # the context is mocked for select_context_class where it is instantiated
    @patch("vantage6.cli.context.NodeContext")
    def test_new_config_already_exists(self, context):
        """No duplicate configurations are allowed."""

        context.config_exists.return_value = True

        runner = CliRunner()
        result = runner.invoke(
            cli_node_new_configuration,
            [
                "--name",
                "some-name",
            ],
        )

        # check that error is produced
        self.assertIn(
            "[error] - Configuration some-name already exists!", result.output
        )

        # check non-zero exit code
        self.assertEqual(result.exit_code, 1)

    @patch("vantage6.cli.common.new.ensure_config_dir_writable")
    @patch("vantage6.cli.node.common.NodeContext")
    def test_new_write_permissions(self, context, permissions):
        """User needs write permissions."""

        context.config_exists.return_value = False
        permissions.return_value = False

        runner = CliRunner()
        result = runner.invoke(
            cli_node_new_configuration,
            [
                "--name",
                "some-name",
            ],
        )

        # check that error is produced
        self.assertTrue("[error] - Your user does not have write" in result.output)

        # check non-zero exit code
        self.assertEqual(result.exit_code, 1)

    @patch("vantage6.cli.common.decorator.get_context")
    @patch("vantage6.cli.common.decorator.select_configuration_questionnaire")
    def test_files(self, select_config, context):
        """No errors produced when retrieving filepaths."""

        context.return_value = MagicMock(
            config_file="/file.yaml", log_file="/log.log", data_dir="/dir"
        )
        context.return_value.databases.items.return_value = [["label", "/file.db"]]
        select_config.return_value = "iknl"

        runner = CliRunner()
        result = runner.invoke(cli_node_files, [])

        # we check that no warnings have been produced
        self.assertEqual(result.output[:7], "[info ]")

        # check status code is OK
        self.assertEqual(result.exit_code, 0)

    @patch("vantage6.cli.node.common.NodeContext")
    def test_files_non_existing_config(self, context):
        """An error is produced when a non existing config is used."""

        context.config_exists.return_value = False

        runner = CliRunner()
        result = runner.invoke(cli_node_files, ["--name", "non-existing"])

        # Check that error is produced
        self.assertEqual(result.output[:7], "[error]")

        # check for non zero exit-code
        self.assertNotEqual(result.exit_code, 0)

    @patch("vantage6.cli.node.start.select_k8s_config")
    @patch("os.makedirs")
    @patch("vantage6.cli.common.decorator.get_context")
    @patch("vantage6.cli.node.start.helm_install")
    @patch("vantage6.cli.node.start.create_task_namespace_if_not_exists")
    def test_start(
        self,
        create_task_namespace_if_not_exists,
        helm_install,
        context,
        os_makedirs,
        ctx_ns,
    ):
        """Start node without errors"""

        ctx = MagicMock(
            config={"node": {"proxyPort": 7654, "image": "my-node-image"}},
            config_file="/config.yaml",
            data_dir=Path("."),
            log_dir=Path("."),
            helm_release_name="vantage6-test-node",
            is_sandbox=False,
        )
        ctx.config_exists.return_value = True
        ctx.name = "test-node"
        context.return_value = ctx
        ctx_ns.return_value = KubernetesConfig(
            context="test-context",
            namespace="test-namespace",
        )

        runner = CliRunner()
        result = runner.invoke(cli_node_start, ["--name", "test-node"])

        self.assertEqual(result.exit_code, 0)

    @patch("vantage6.cli.common.stop.select_k8s_config")
    @patch("vantage6.cli.common.stop.get_context")
    @patch("vantage6.cli.common.utils.find_running_service_names")
    @patch("vantage6.cli.node.stop._stop_node")
    def test_stop(self, _stop_node, find_running_service_names, get_context, ctx_ns):
        node_name = "iknl"
        node_helm_name = f"{APPNAME}-{node_name}-user-node"
        find_running_service_names.return_value = [node_helm_name]
        get_context.return_value = MagicMock(helm_release_name=node_helm_name)
        ctx_ns.return_value = KubernetesConfig(
            context="test-context",
            namespace="test-namespace",
        )

        runner = CliRunner()
        result = runner.invoke(cli_node_stop, ["--name", "iknl"])

        self.assertEqual(result.exit_code, 0)
        self.assertIsNone(result.exception)

    @patch("vantage6.cli.common.stop.select_k8s_config")
    @patch("vantage6.cli.node.restart.select_k8s_config")
    @patch("vantage6.cli.node.restart.subprocess.run")
    @patch("vantage6.cli.common.stop.get_context")
    @patch("vantage6.cli.node.stop._stop_node")
    def test_restart(
        self, _stop_node, get_context, subprocess_run, ctx_ns_stop, ctx_ns_restart
    ):
        """Restart a node without errors."""
        node_name = "iknl"
        node_helm_name = f"{APPNAME}-{node_name}-user-node"
        get_context.return_value = MagicMock(helm_release_name=node_helm_name)
        ctx_ns_stop.return_value = KubernetesConfig(
            context="test-context",
            namespace="test-namespace",
        )
        ctx_ns_restart.return_value = KubernetesConfig(
            context="test-context",
            namespace="test-namespace",
        )

        # Mock subprocess.run to handle both helm calls and node start calls
        def mock_subprocess_run(*args, **kwargs):
            # Check if this is a helm command (contains 'helm list')
            if "helm" in args[0] and "list" in args[0]:
                return MagicMock(
                    stdout='[{"name": "' + node_helm_name + '", "status": "deployed"}]',
                    returncode=0,
                )
            else:
                # For other subprocess calls (like starting the node)
                return MagicMock(returncode=0)

        subprocess_run.side_effect = mock_subprocess_run
        runner = CliRunner()
        result = runner.invoke(cli_node_restart, ["--name", "iknl"])
        self.assertEqual(result.exit_code, 0)

    @patch("vantage6.cli.node.attach.attach_logs")
    @patch("vantage6.cli.node.attach.select_k8s_config")
    def test_attach(self, select_k8s_config, attach_logs):
        """Attach pod logs without errors."""
        select_k8s_config.return_value = KubernetesConfig(
            context="test-context",
            namespace="test-namespace",
        )
        runner = CliRunner()
        runner.invoke(cli_node_attach)
        attach_logs.assert_called_once_with(
            name=None,
            instance_type=InstanceType.NODE,
            infra_component=InfraComponentName.NODE,
            system_folders=False,
            k8s_config=KubernetesConfig(
                context="test-context",
                namespace="test-namespace",
            ),
            is_sandbox=False,
        )

    @patch("vantage6.cli.node.create_private_key.create_client_and_authenticate")
    @patch("vantage6.cli.node.common.NodeContext")
    @patch("vantage6.cli.node.create_private_key.NodeContext")
    def test_create_private_key(self, context, common_context, client):
        common_context.config_exists.return_value = True
        context.return_value.type_data_folder.return_value = Path(".")
        client.return_value = MagicMock(whoami=MagicMock(organization_name="Test"))
        # client.whoami.organization_name = "Test"
        runner = CliRunner()

        result = runner.invoke(cli_node_create_private_key, ["--name", "application"])

        self.assertEqual(result.exit_code, 0)

        # remove the private key file again
        os.remove("privkey_Test.pem")

    @patch("vantage6.cli.node.create_private_key.RSACryptor")
    @patch("vantage6.cli.node.create_private_key.create_client_and_authenticate")
    @patch("vantage6.cli.node.common.NodeContext")
    @patch("vantage6.cli.node.create_private_key.NodeContext")
    def test_create_private_key_overwite(
        self, context, common_context, client, cryptor
    ):
        common_context.config_exists.return_value = True
        context.return_value.type_data_folder.return_value = Path(".")
        client.return_value = MagicMock(whoami=MagicMock(organization_name="Test"))
        cryptor.create_public_key_bytes.return_value = b""
        # client.whoami.organization_name = "Test"

        runner = CliRunner()

        # overwrite
        with runner.isolated_filesystem():
            with open("privkey_iknl.pem", "w") as f:
                f.write("does-not-matter")

            result = runner.invoke(
                cli_node_create_private_key,
                ["--name", "application", "--overwrite", "--organization-name", "iknl"],
            )
        self.assertEqual(result.exit_code, 0)

        # do not overwrite
        with runner.isolated_filesystem():
            with open("privkey_iknl.pem", "w") as f:
                f.write("does-not-matter")

            result = runner.invoke(
                cli_node_create_private_key,
                ["--name", "application", "--organization-name", "iknl"],
            )

        self.assertEqual(result.exit_code, 0)

    @patch("vantage6.cli.node.common.NodeContext")
    def test_create_private_key_config_not_found(self, context):
        context.config_exists.return_value = False

        runner = CliRunner()
        result = runner.invoke(cli_node_create_private_key, ["--name", "application"])

        self.assertEqual(result.exit_code, 1)

    @patch("vantage6.cli.node.common.info")
    @patch("vantage6.cli.node.common.debug")
    @patch("vantage6.cli.node.common.error")
    @patch("vantage6.cli.node.common.UserClient")
    def test_client(self, client, error, debug, info):
        ctx = MagicMock(
            config={
                "node": {
                    "hq": {
                        "url": LOCALHOST,
                        "port": Ports.DEV_HQ.value,
                        "path": "",
                    }
                }
            }
        )

        # should not trigger an exception
        try:
            create_client_and_authenticate(ctx)
        except Exception:
            self.fail("Raised an exception!")

        # client raises exception
        client.side_effect = Exception("Boom!")
        with self.assertRaises(Exception):
            create_client_and_authenticate(ctx)



import os
import sys
from typing import Optional

import click
import requests

from vantage6.common import error, info, warning

from vantage6.cli.auth.install import check_and_install_keycloak_operator
from vantage6.cli.common.k8s_utils import run_kubectl_command
from vantage6.cli.k8s_config import KubernetesConfig, select_k8s_config


def _get_latest_cert_manager_version() -> str:
    """
    Get the latest cert-manager release tag from GitHub.

    Raises a RuntimeError if the lookup fails, so that the caller can instruct
    the user to provide the version explicitly.
    """
    github_api_url = (
        "https://api.github.com/repos/cert-manager/cert-manager/releases/latest"
    )

    headers = {"Accept": "application/vnd.github.v3+json"}
    github_token = os.environ.get("GITHUB_TOKEN")
    if github_token:
        headers["Authorization"] = f"token {github_token}"

    try:
        response = requests.get(github_api_url, timeout=5, headers=headers)
        response.raise_for_status()
        release_data = response.json()
        version = release_data.get("tag_name")
        if not version:
            raise RuntimeError(
                "No tag_name found in cert-manager GitHub releases response."
            )
        return version
    except Exception as exc:  # pragma: no cover - best-effort helper
        raise RuntimeError(
            f"Failed to fetch latest cert-manager version from GitHub: {exc!r}"
        ) from exc


def _check_cert_manager_crds_installed(k8s_config: KubernetesConfig) -> bool:
    """
    Check if the cert-manager Certificate CRD is already installed.

    We only check for the `certificates.cert-manager.io` CRD here, as that is what
    the hub chart requires to create Certificate resources.
    """
    try:
        result = run_kubectl_command(
            ["get", "crd", "certificates.cert-manager.io"],
            k8s_config=k8s_config,
            check=False,
            use_k8s_config_namespace=False,
        )
        return result.returncode == 0
    except Exception:
        return False


def check_and_install_cert_manager_crds(
    k8s_config: KubernetesConfig,
    version: Optional[str] = None,
) -> None:
    """
    Ensure that cert-manager CRDs are installed in the cluster.

    This installs only the CRDs required for the hub chart (in particular
    `certificates.cert-manager.io`) so that Certificate resources can be created.
    """
    try:
        if _check_cert_manager_crds_installed(k8s_config):
            info("✅ cert-manager CRDs are already installed.")
            return

        info("⚠️  cert-manager CRDs are not installed.")
        info("Installing cert-manager CRDs...")

        if version is None:
            try:
                version = _get_latest_cert_manager_version()
            except RuntimeError as exc:
                error(str(exc))
                error(
                    "Please re-run 'v6 hub install' with the "
                    "--cert-manager-version flag to specify the cert-manager "
                    "version explicitly (for example: 'v1.15.3')."
                )
                sys.exit(1)

        crds_url = (
            "https://github.com/cert-manager/cert-manager/releases/download/"
            f"{version}/cert-manager.crds.yaml"
        )
        run_kubectl_command(
            ["apply", "-f", crds_url],
            k8s_config=k8s_config,
        )
        info("✅ cert-manager CRDs installed successfully.")
    except Exception as exc:  # pragma: no cover - best-effort helper
        warning(f"⚠️  Could not ensure cert-manager CRDs are installed: {exc}")
        warning(
            "If hub installation fails due to missing Certificate CRDs, please run "
            "'v6 hub install' or install cert-manager manually and try again."
        )


def cert_manager_seems_installed(k8s_config: KubernetesConfig) -> bool:
    """
    Heuristically check whether cert-manager (or an equivalent controller)
    appears to be installed.

    We do not attempt to install cert-manager automatically from the hub CLI
    because clusters may already have a controller that owns the
    'webhook.cert-manager.io' ValidatingWebhookConfiguration (for example via a
    platform component such as 'admissionsenforcer').
    """
    # Primary check: a cert-manager deployment in the cert-manager namespace.
    try:
        result = run_kubectl_command(
            ["get", "deployment", "cert-manager", "-n", "cert-manager"],
            k8s_config=k8s_config,
            check=False,
            # intentionally not using the namespace from the Kubernetes configuration
            use_k8s_config_namespace=False,
        )
        if result.returncode == 0:
            return True
    except Exception:
        pass

    # Fallback: presence of the webhook configuration is a strong indicator
    # that something is already managing cert-manager-style webhooks.
    try:
        result = run_kubectl_command(
            ["get", "validatingwebhookconfiguration", "webhook.cert-manager.io"],
            k8s_config=k8s_config,
            check=False,
        )
        return result.returncode == 0
    except Exception:
        return False


@click.command()
@click.option("--context", default=None, help="Kubernetes context to use")
@click.option("--namespace", default=None, help="Kubernetes namespace to use")
@click.option(
    "--cert-manager-version",
    default=None,
    help=(
        "cert-manager version tag to use for CRDs. "
        "If omitted, the latest release is looked up from GitHub; "
        "if that lookup fails, you will be asked to provide this flag explicitly."
    ),
)
def cli_hub_install(
    context: str | None,
    namespace: str | None,
    cert_manager_version: str | None,
) -> None:
    """
    Install prerequisites for running a vantage6 hub.

    This command installs:

    - cert-manager CRDs (so that Certificate resources can be created)
    - the Keycloak operator (and its CRDs), reused from `v6 auth install-keycloak`
    """
    k8s_config = select_k8s_config(context=context, namespace=namespace)

    # Ensure cert-manager CRDs are present.
    check_and_install_cert_manager_crds(k8s_config, version=cert_manager_version)

    # Ensure Keycloak operator (and its CRDs) are present.
    try:
        check_and_install_keycloak_operator(k8s_config)
    except SystemExit:
        # Bubble up a clear error instead of a silent exit.
        error(
            "Failed to install the Keycloak operator while running 'v6 hub install'. "
            "Please inspect the logs above and try again."
        )
        exit(1)



from vantage6.common.enum import EnumBase


class AuthCredentials(EnumBase):
    KEYCLOAK_ADMIN_USER = ("keycloakAdminUser", "Keycloak admin username")
    KEYCLOAK_ADMIN_PASSWORD = ("keycloakAdminPassword", "Keycloak admin password")
    VANTAGE6_ADMIN_PASSWORD = ("vantage6AdminPassword", "Vantage6 admin password")

    def __new__(cls, value: str, description: str):
        # EnumBase inherits from Enum; calling its __new__ triggers a guard in
        # the stdlib Enum that raises TypeError. Bypass Enum's __new__ and
        # use object.__new__ directly.
        obj = object.__new__(cls)
        obj._value_ = value
        obj.description = description
        return obj



import click

from vantage6.common.globals import InstanceType

from vantage6.cli.common.list import get_configuration_list


@click.command()
def cli_node_list() -> None:
    """
    Lists all node configurations.

    Note that this command cannot find node configuration files in custom
    directories.
    """
    get_configuration_list(InstanceType.NODE)



import subprocess
import sys

import click

from vantage6.common import error, info

from vantage6.cli.dev.common import check_devspace_installed


@click.command()
def cli_clean_dev_env():
    """
    Stops and cleans up the development environment.

    Removes the kubernetes resources and local data (e.g. tasks data, database data).
    This is useful when you want to start from scratch.
    """
    check_devspace_installed()

    try:
        info("🧹 Cleaning development environment with devspace...")
        cmd = ["devspace", "run", "purge"]
        subprocess.run(cmd, check=True, capture_output=False)
        info("✅ Development environment cleaned successfully!")
    except subprocess.CalledProcessError as e:
        error(f"❌ Error cleaning development environment: {e}")
        sys.exit(e.returncode)



import itertools
from pathlib import Path
from shutil import rmtree

import click

from vantage6.common import error, warning
from vantage6.common.globals import InstanceType

from vantage6.cli.common.remove import execute_remove
from vantage6.cli.configuration_create import select_configuration_questionnaire
from vantage6.cli.context import get_context
from vantage6.cli.context.node import NodeContext
from vantage6.cli.globals import InfraComponentName
from vantage6.cli.hub.remove import cli_hub_remove


@click.command()
@click.option("-n", "--name", default=None, help="Name of the configuration.")
@click.option(
    "--data-dir",
    "custom_data_dir",
    type=click.Path(exists=True),
    default=None,
    help="Path to a custom data directory to use. This option is especially useful "
    "on WSL because of mount issues for default directories. Use the same value as "
    "was provided when creating the sandbox.",
)
@click.pass_context
def cli_sandbox_remove(
    click_ctx: click.Context,
    name: str | None,
    custom_data_dir: Path | None,
) -> None:
    """Remove all related demo network files and folders.

    Select a hub configuration to remove. The related sandbox components will also be
    removed.
    """

    if not name:
        try:
            name = select_configuration_questionnaire(
                type_=InstanceType.HUB, system_folders=False, is_sandbox=True
            )
        except Exception:
            error("No configurations could be found!")
            exit()

    ctx = get_context(InstanceType.HUB, name, system_folders=False, is_sandbox=True)

    # remove the nodes
    NodeContext.LOGGING_ENABLED = False
    configs, _ = NodeContext.available_configurations(system_folders=False)
    node_names = [
        config.name for config in configs if config.name.startswith(f"{ctx.name}-node-")
    ]
    for name in node_names:
        # Context clases are singletons, so we need to clear the cache to force
        # creation of a new instance. Otherwise, we *think* we get the ctx of another
        # node but we actually get the one from the previous node in the loop.
        if NodeContext in NodeContext._instances:
            del NodeContext._instances[NodeContext]

        node_ctx = NodeContext(
            instance_name=name,
            system_folders=False,
            is_sandbox=True,
            print_log_header=False,
            logger_prefix="",
            in_container=False,
        )
        for handler in itertools.chain(
            node_ctx.log.handlers, node_ctx.log.root.handlers
        ):
            handler.close()
        execute_remove(
            node_ctx,
            InstanceType.NODE,
            InfraComponentName.NODE,
            name,
            system_folders=False,
            force=True,
        )

    # remove data files attached to the network
    data_dirs_nodes = NodeContext.instance_folders("node", "", False)["dev"]
    try:
        rmtree(Path(data_dirs_nodes / ctx.name))
    except Exception as e:
        warning(f"Failed to delete data directory {data_dirs_nodes / ctx.name}: {e}")

    # remove the hub last - if anything goes wrong, the hub is still there so the
    # user can still retry the removal.
    # Note that this also checks if the hub is running. Therefore, it is prevented
    # that a running sandbox is removed.
    for handler in itertools.chain(ctx.log.handlers, ctx.log.root.handlers):
        handler.close()
    click_ctx.invoke(
        cli_hub_remove, ctx=ctx, name=name, system_folders=False, force=True
    )

    # remove the right data in the custom data directory if it is provided. If a custom
    # data directory is provided, all files related to the sandbox have been stored in
    # a subfolder with the name of the sandbox.
    if custom_data_dir:
        sandbox_data_dir = custom_data_dir / ctx.name
        if sandbox_data_dir.is_dir():
            rmtree(sandbox_data_dir)



"""
This module contains global variables that are used throughout the CLI.
"""

from pathlib import Path

from vantage6.common.enum import StrEnumBase
from vantage6.common.globals import APPNAME

#
#  CLI SETTINGS
#

DEFAULT_CLI_CONFIG_FILE = "kubernetes_config.yaml"

#
#   BACKEND SETTINGS
#
DEFAULT_API_SERVICE_SYSTEM_FOLDERS = True

#
#   NODE SETTINGS
#
DEFAULT_NODE_SYSTEM_FOLDERS = False

#
#   INSTALLATION SETTINGS
#
PACKAGE_FOLDER = Path(__file__).parent.parent.parent

# FIXME BvB 22-06-28 think this is also defined in the node globals, and this
# one appears not to be used
# TODO Check and remove
# HCR: This constants is indeed not used, as well as the other defined in node/globals.py
# NODE_V6_PROXY_HOSTNAME = "proxyserver"

DATA_FOLDER = PACKAGE_FOLDER / APPNAME / "_data"

# with open(Path(PACKAGE_FOLDER) / APPNAME / "cli" / "VERSION") as f:
#     VERSION = f.read()

# Maximum time to start up RabbitMQ in seconds
RABBIT_TIMEOUT = 300

# Location of repository to create new algorithm templates from
ALGORITHM_TEMPLATE_REPO = "gh:vantage6/v6-algorithm-template.git"

# image to use for diagnostics in `v6 test` commands
DIAGNOSTICS_IMAGE = "harbor2.vantage6.ai/algorithms/diagnostic:uluru"

# Address of community algorithm store
COMMUNITY_STORE = "https://store.uluru.vantage6.ai/api"

DEFAULT_PROMETHEUS_IMAGE = "prom/prometheus"
PROMETHEUS_CONFIG = "prometheus.yml"
PROMETHEUS_DIR = "prometheus"

# template configuration files
TEMPLATE_FOLDER = PACKAGE_FOLDER / APPNAME / "cli" / "template"
HQ_TEMPLATE_FILE = "hq_config.j2"
NODE_TEMPLATE_FILE = "node_config.j2"
ALGO_STORE_TEMPLATE_FILE = "algo_store_config.j2"
AUTH_TEMPLATE_FILE = "auth_config.j2"
HUB_TEMPLATE_FILE = "hub_config.j2"


# datasets included in the nodes of the dev network
class DefaultDatasets(StrEnumBase):
    """Enum containing default datasets"""

    OLYMPIC_ATHLETES = "olympic_athletes_2016.csv"
    KAPLAN_MEIER_TEST = "km_dataset.csv"


class BackendType(StrEnumBase):
    """Enum containing backend types"""

    HQ = "hq"
    ALGORITHM_STORE = "algorithm-store"


class ChartName(StrEnumBase):
    """Enum containing chart names"""

    HQ = "hq"
    ALGORITHM_STORE = "store"
    NODE = "node"
    AUTH = "auth"
    HUB = "hub"


class CLICommandName(StrEnumBase):
    """Enum containing CLI command names"""

    HQ = "hq"
    ALGORITHM_STORE = "algorithm-store"
    NODE = "node"
    ALGORITHM = "algorithm"
    TEST = "test"
    SANDBOX = "sandbox"
    DEV = "dev"
    USE = "use"
    AUTH = "auth"
    HUB = "hub"


class InfraComponentName(StrEnumBase):
    """Enum containing infrastructure components"""

    HQ = "hq"
    ALGORITHM_STORE = "store"
    NODE = "node"
    AUTH = "auth"
    HUB = "hub"



import click

from vantage6.common import info
from vantage6.common.globals import InstanceType

from vantage6.cli.common.decorator import click_insert_context
from vantage6.cli.context.algorithm_store import AlgorithmStoreContext


@click.command()
@click.option("--sandbox/--no-sandbox", "sandbox", default=False)
@click_insert_context(type_=InstanceType.ALGORITHM_STORE, sandbox_param="sandbox")
def cli_algo_store_files(ctx: AlgorithmStoreContext) -> None:
    """
    List files that belong to a particular algorithm store instance.
    """
    info(f"Configuration file = {ctx.config_file}")
    info(f"Log file           = {ctx.log_file}")
    info(f"Database           = {ctx.get_database_uri()}")



import unittest
from pathlib import Path
from unittest.mock import MagicMock, patch

from click.testing import CliRunner

from vantage6.common.globals import APPNAME, InstanceType

from vantage6.cli.common.attach import attach_logs
from vantage6.cli.globals import InfraComponentName
from vantage6.cli.hq.attach import cli_hq_attach
from vantage6.cli.hq.files import cli_hq_files
from vantage6.cli.hq.import_ import cli_hq_import
from vantage6.cli.hq.list import cli_hq_configuration_list
from vantage6.cli.hq.start import cli_hq_start
from vantage6.cli.hq.stop import cli_hq_stop
from vantage6.cli.k8s_config import KubernetesConfig


class CLITestHQ(unittest.TestCase):
    @patch("vantage6.cli.hq.start.select_k8s_config")
    @patch("os.makedirs")
    @patch("vantage6.cli.common.decorator.get_context")
    @patch("vantage6.cli.hq.start.helm_install")
    def test_start(
        self,
        context,
        helm_install,
        os_makedirs,
        ctx_ns,
    ):
        """Start HQ without errors"""

        ctx = MagicMock(
            config={"uri": "sqlite:///file.db", "port": 9999},
            config_file="/config.yaml",
            data_dir=Path("."),
        )
        ctx.config_exists.return_value = True
        ctx.name = "not-running"
        context.return_value = ctx
        ctx_ns.return_value = KubernetesConfig(
            context="test-context",
            namespace="test-namespace",
        )

        runner = CliRunner()
        result = runner.invoke(cli_hq_start, ["--name", "not-running"])

        self.assertEqual(result.exit_code, 0)

    @patch("vantage6.cli.context.hq.HQContext.available_configurations")
    @patch("vantage6.cli.common.list.find_running_service_names")
    def test_configuration_list(
        self, find_running_service_names, available_configurations
    ):
        """Configuration list without errors."""
        hq_name = "iknl"
        find_running_service_names.return_value = [
            f"{APPNAME}-{hq_name}-system-{InstanceType.HQ.value}"
        ]

        # returns a list of configurations and failed inports
        def side_effect(system_folders):
            config = MagicMock()
            config.name = hq_name
            if not system_folders:
                return [[config], []]
            else:
                return [[config], []]

        available_configurations.side_effect = side_effect

        runner = CliRunner()
        result = runner.invoke(cli_hq_configuration_list)

        self.assertEqual(result.exit_code, 0)
        self.assertIsNone(result.exception)

    @patch("vantage6.cli.common.decorator.get_context")
    def test_files(self, context):
        """Configuration files without errors."""

        ctx = context.return_value = MagicMock(
            log_file="/log_file.log", config_file="/iknl.yaml"
        )
        ctx.get_database_uri.return_value = "sqlite:///test.db"

        runner = CliRunner()
        result = runner.invoke(cli_hq_files, ["--name", "iknl"])

        self.assertIsNone(result.exception)
        self.assertEqual(result.exit_code, 0)

    @patch("vantage6.cli.hq.import_.UserClient")
    @patch("vantage6.cli.hq.import_.requests.get")
    @patch("vantage6.cli.hq.import_.click.Path")
    @patch("vantage6.cli.common.decorator.get_context")
    def test_import(self, context, click_path, requests_get, user_client):
        """Import entities without errors."""
        click_path.return_value = MagicMock()
        requests_get.return_value = MagicMock(
            status_code=200, json=lambda: {"version": "1.0.0"}
        )
        user_client.return_value = MagicMock()
        user_client.authenticate.return_value = True

        ctx = MagicMock()
        ctx.name = "some-name"
        context.return_value = ctx

        # Mock the version to match the hq version
        with patch("vantage6.cli.hq.import_.__version__", "1.0.0"):
            runner = CliRunner()
            with runner.isolated_filesystem():
                with open("some.yaml", "w") as fp:
                    fp.write("""
organizations:
  - name: Test Org
    address1: Test Address
    zipcode: 12345
    country: Test Country
    domain: test.com
    public_key: test-key
    users: []
collaborations: []
""")
                result = runner.invoke(cli_hq_import, ["--name", "iknl", "some.yaml"])

        self.assertIsNone(result.exception)
        self.assertEqual(result.exit_code, 0)

        # check that the import fails when the yaml is invalid
        with patch("vantage6.cli.hq.import_.__version__", "1.0.0"):
            runner = CliRunner()
            with runner.isolated_filesystem():
                with open("some.yaml", "w") as fp:
                    fp.write("invalid")
                result = runner.invoke(cli_hq_import, ["--name", "iknl", "some.yaml"])
                self.assertEqual(result.exit_code, 1)
                self.assertIsNotNone(result.exception)

    @patch("vantage6.cli.common.stop.select_k8s_config")
    @patch("vantage6.cli.common.stop.find_running_service_names")
    @patch("vantage6.cli.common.stop.get_context")
    @patch("vantage6.cli.hq.stop._stop_hq")
    def test_stop(
        self,
        _stop_hq,
        get_context,
        find_running_service_names,
        context_and_namespace,
    ):
        """Stop HQ without errors."""

        instance_name = "iknl"
        hq_name = f"{APPNAME}-{instance_name}-system-{InstanceType.HQ.value}"

        find_running_service_names.return_value = [hq_name]
        get_context.return_value = MagicMock(helm_release_name=hq_name)
        context_and_namespace.return_value = KubernetesConfig(
            context="test-context",
            namespace="test-namespace",
        )

        runner = CliRunner()
        result = runner.invoke(cli_hq_stop, ["--name", instance_name])

        self.assertIsNone(result.exception)
        self.assertEqual(result.exit_code, 0)

    @patch("vantage6.cli.hq.attach.attach_logs")
    @patch("vantage6.cli.hq.attach.select_k8s_config")
    def test_attach(self, select_k8s_config, attach_logs):
        """Attach logs to the console without errors."""
        select_k8s_config.return_value = KubernetesConfig(
            context="test-context",
            namespace="test-namespace",
        )
        runner = CliRunner()
        result = runner.invoke(cli_hq_attach)

        self.assertIsNone(result.exception)
        self.assertEqual(result.exit_code, 0)
        attach_logs.assert_called_once_with(
            name=None,
            instance_type=InstanceType.HQ,
            infra_component=InfraComponentName.HQ,
            system_folders=False,
            k8s_config=KubernetesConfig(
                context="test-context",
                namespace="test-namespace",
            ),
            is_sandbox=False,
            additional_labels="component=vantage6-hq",
        )

    @patch("vantage6.cli.common.utils.subprocess.run")
    @patch("vantage6.cli.common.attach.select_running_service")
    @patch("vantage6.cli.hq.attach.select_k8s_config")
    @patch("vantage6.cli.common.attach.Popen")
    def test_attach_logs(
        self,
        mock_popen,
        select_k8s_config,
        select_running_service,
        subprocess_run,
    ):
        select_k8s_config.return_value = KubernetesConfig(
            context="test-context",
            namespace="test-namespace",
        )
        select_running_service.return_value = "vantage6-iknl-system-hq"

        # Mock subprocess.run to return success for helm list command
        subprocess_run.return_value = MagicMock(
            returncode=0,
            stdout='[{"name": "vantage6-iknl-system-hq", "status": "deployed"}]',
        )

        # Mock the Popen instance and its methods
        mock_process = mock_popen.return_value
        mock_process.wait.return_value = None

        # Call the function with a sample label
        attach_logs(
            name=None,
            instance_type=InstanceType.HQ,
            infra_component=InfraComponentName.HQ,
            system_folders=True,
            k8s_config=select_k8s_config.return_value,
            is_sandbox=False,
            additional_labels="test=label",
        )

        # Construct the expected command
        expected_command = [
            "kubectl",
            "--context",
            "test-context",
            "-n",
            "test-namespace",
            "logs",
            "--follow",
            "--selector",
            "release=vantage6-iknl-system-hq,test=label",
            "--all-containers=true",
        ]

        # Verify that Popen was called with the expected command
        mock_popen.assert_called_once_with(expected_command, stdout=None, stderr=None)

        # Verify that wait was called on the process
        mock_process.wait.assert_called_once()



from pathlib import Path

import click
from colorama import Fore, Style

from vantage6.common import error
from vantage6.common.globals import InstanceType

from vantage6.cli.configuration_create import get_external_database_url
from vantage6.cli.context import get_context
from vantage6.cli.context.hub import HubContext
from vantage6.cli.k8s_config import select_k8s_config
from vantage6.cli.sandbox.config.hub import SandboxHubConfigManager
from vantage6.cli.sandbox.start import execute_sandbox_start
from vantage6.cli.utils import prompt_config_name


@click.command()
@click.option(
    "-n", "--name", default=None, type=str, help="Name for your development setup"
)
@click.option(
    "--num-nodes",
    type=int,
    default=3,
    help="Generate this number of nodes in the development network",
)
@click.option(
    "--hq-image",
    type=str,
    default=None,
    help="HQ docker image to use when setting up resources",
)
@click.option(
    "--ui-image",
    type=str,
    default=None,
    help="UI docker image to specify in configuration files. Will be used on startup of"
    " the network",
)
@click.option(
    "--store-image",
    type=str,
    default=None,
    help="Algorithm store docker image to use when setting up resources for "
    "the development algorithm store",
)
@click.option(
    "--node-image",
    type=str,
    default=None,
    help="Node docker image to use when setting up resources for the development node",
)
@click.option(
    "--extra-hq-config",
    type=click.Path(exists=True),
    default=None,
    help="YAML File with additional HQ configuration. This will be appended to the HQ "
    "configuration file",
)
@click.option(
    "--extra-node-config",
    type=click.Path("rb"),
    default=None,
    help="YAML File with additional node configuration. This will be"
    " appended to each of the node configuration files",
)
@click.option(
    "--extra-store-config",
    type=click.Path("rb"),
    default=None,
    help="YAML File with additional algorithm store configuration. This will be"
    " appended to the algorithm store configuration file",
)
@click.option(
    "--extra-auth-config",
    type=click.Path("rb"),
    default=None,
    help="YAML File with additional auth configuration. This will be"
    " appended to the auth configuration file",
)
@click.option(
    "--add-dataset",
    type=(str, click.Path()),
    default=None,
    multiple=True,
    help="Add a dataset to the nodes. The first argument is the label of the database, "
    "the second is the path to the dataset file.",
)
@click.option("--context", default=None, help="Kubernetes context to use")
@click.option("--namespace", default=None, help="Kubernetes namespace to use")
@click.option(
    "--data-dir",
    type=click.Path(exists=True),
    default=None,
    help="Path to a custom data directory to use. This option is especially useful "
    "on WSL because of mount issues for default directories",
)
@click.option(
    "--with-prometheus", is_flag=True, default=False, help="Enable Prometheus"
)
@click.option(
    "--local-chart-dir",
    type=click.Path(exists=True),
    default=None,
    help="Local chart repository to use.",
)
@click.option(
    "--external-dbs",
    is_flag=True,
    default=False,
    help="Use external databases instead of deploying internal ones. You will be "
    "prompted for database URIs for auth, HQ, and algorithm store.",
)
def cli_new_sandbox(
    name: str,
    num_nodes: int,
    hq_image: str | None,
    ui_image: str | None,
    store_image: str | None,
    node_image: str | None,
    extra_hq_config: Path | None,
    extra_node_config: Path | None,
    extra_store_config: Path | None,
    extra_auth_config: Path | None,
    add_dataset: tuple[str, Path] | None,
    context: str | None,
    namespace: str | None,
    data_dir: str | None,
    local_chart_dir: Path | None,
    with_prometheus: bool,
    external_dbs: bool,
) -> None:
    """
    Create a sandbox environment.
    """

    # Prompt for the k8s namespace and context
    k8s_config = select_k8s_config(context=context, namespace=namespace)

    if data_dir is not None:
        data_dir = Path(data_dir)
        if not data_dir.exists():
            error(f"Data directory {data_dir} does not exist!")
            exit(1)

    hub_name = prompt_config_name(name)
    if HubContext.config_exists(hub_name, False, is_sandbox=True):
        error(f"Configuration {Fore.RED}{hub_name}{Style.RESET_ALL} already exists!")
        exit(1)

    # If using external databases, prompt for database URIs
    external_db_uris = None
    if external_dbs:
        auth_uri = get_external_database_url(InstanceType.AUTH)
        hq_uri = get_external_database_url(InstanceType.HQ)
        store_uri = get_external_database_url(InstanceType.ALGORITHM_STORE)
        external_db_uris = {"auth": auth_uri, "hq": hq_uri, "store": store_uri}

    sb_config_manager = SandboxHubConfigManager(
        hub_name=hub_name,
        hq_image=hq_image,
        ui_image=ui_image,
        store_image=store_image,
        extra_hq_config=extra_hq_config,
        extra_store_config=extra_store_config,
        extra_auth_config=extra_auth_config,
        k8s_config=k8s_config,
        with_prometheus=with_prometheus,
        custom_data_dir=data_dir,
        external_db_uris=external_db_uris,
    )

    sb_config_manager.generate_hub_configs()

    ctx = get_context(
        type_=InstanceType.HUB,
        name=hub_name,
        system_folders=False,
        is_sandbox=True,
    )

    execute_sandbox_start(
        ctx=ctx,
        hub_name=hub_name,
        k8s_config=k8s_config,
        num_nodes=num_nodes,
        initialize=True,
        node_image=node_image,
        extra_node_config=extra_node_config,
        add_dataset=add_dataset,
        custom_data_dir=data_dir,
        local_chart_dir=local_chart_dir,
    )



import subprocess
import time

from vantage6.common import info, warning

from vantage6.cli.common.k8s_utils import run_kubectl_command
from vantage6.cli.k8s_config import KubernetesConfig


def wait_for_keycloak_ready(release_name: str, k8s_config: KubernetesConfig) -> None:
    """
    Ensure Keycloak pod is ready and realm import job finished.

    Parameters
    ----------
    release_name : str
        The name of the release.
    k8s_config : KubernetesConfig
        The Kubernetes configuration to use.
    """
    selector = f"app.kubernetes.io/instance={release_name}-kc"
    job_name = f"{release_name}-realm-import"

    info("Waiting for Keycloak pod to be created...")
    while True:
        result = run_kubectl_command(
            ["get", "pod", "-l", selector, "-o", "name"],
            k8s_config,
            check=False,
        )
        if result.stdout.strip():
            break
        time.sleep(1)
    info("Keycloak pod was created, waiting for it to be ready...")
    run_kubectl_command(
        ["wait", "--for=condition=ready", "pod", "-l", selector, "--timeout", "300s"],
        k8s_config,
    )

    info("Waiting for Keycloak realm import job to be created...")
    while True:
        result = run_kubectl_command(["get", "job", job_name], k8s_config, check=False)
        if result.returncode == 0:
            break
        else:
            time.sleep(1)
    info("Keycloak realm import job was created, waiting for it to finish...")
    # Check if the job exists first (don't raise exception if it doesn't)
    job_check = run_kubectl_command(["get", "job", job_name], k8s_config, check=False)
    if job_check.returncode == 0:
        # Job exists, wait for it to complete
        try:
            run_kubectl_command(
                [
                    "wait",
                    "--for=condition=complete",
                    f"job/{job_name}",
                    "--timeout",
                    "120s",
                ],
                k8s_config,
            )
            info("Realm import job completed successfully.")
        except subprocess.CalledProcessError:
            warning("Realm import job did not complete in time; continuing startup.")
    else:
        info("No realm import job found (realm import may be disabled).")



import json
import secrets
import string
import subprocess
from pathlib import Path

import questionary as q
from kubernetes import client as k8s_client

from vantage6.common import error
from vantage6.common.globals import (
    APPNAME,
    SANDBOX_SUFFIX,
    InstanceType,
)

from vantage6.cli.globals import CLICommandName
from vantage6.cli.k8s_config import KubernetesConfig
from vantage6.cli.utils import validate_input_cmd_args


def create_directory_if_not_exists(directory: Path) -> None:
    """
    Create a directory.
    """
    try:
        directory.mkdir(parents=True, exist_ok=True)
    except Exception as e:
        error(f"Failed to create directory {directory}: {e}")
        exit(1)


def find_running_service_names(
    instance_type: InstanceType,
    only_system_folders: bool = False,
    only_user_folders: bool = False,
    k8s_config: KubernetesConfig | None = None,
) -> list[str]:
    """
    List running vantage6 infrastructure components.

    Parameters
    ----------
    instance_type : InstanceType
        The type of instance to find running services for
    only_system_folders : bool, optional
        Whether to look for system-based services or not. By default False.
    only_user_folders : bool, optional
        Whether to look for user-based services or not. By default False.
    k8s_config : KubernetesConfig, optional
        The Kubernetes configuration to use. If None, the default Kubernetes
        configuration will be used.

    Returns
    -------
    list[str]
        List of release names that are running
    """
    # Input validation
    if k8s_config:
        validate_input_cmd_args(
            k8s_config.context, "k8s_config.context", allow_none=True
        )
        validate_input_cmd_args(
            k8s_config.namespace, "k8s_config.namespace", allow_none=True
        )
    validate_input_cmd_args(instance_type, "instance type", allow_none=False)
    if only_system_folders and only_user_folders:
        error("Cannot use both only_system_folders and only_user_folders")
        exit(1)

    # Create the command
    command = [
        "helm",
        "list",
        "--output",
        "json",  # Get structured output
    ]

    if k8s_config and k8s_config.context:
        command.extend(["--kube-context", k8s_config.context])

    if k8s_config and k8s_config.namespace:
        command.extend(["--namespace", k8s_config.namespace])
    else:
        command.extend(["--all-namespaces"])

    try:
        result = subprocess.run(
            command,
            capture_output=True,
            text=True,
            check=True,
        )
    except subprocess.CalledProcessError as e:
        error(f"Failed to list Helm releases: {e}")
        return []
    except FileNotFoundError:
        error(
            "Helm command not found. Please ensure Helm is installed and available in "
            "the PATH."
        )
        return []

    try:
        releases = json.loads(result.stdout)
    except json.JSONDecodeError:
        error("Failed to parse Helm output as JSON")
        return []

    # filter services for the vantage6 services that are sought. These have
    # the following pattern:
    # f"{APPNAME}-{name}-{scope}-{instance_type.value}"

    # filter for the instance type
    svc_starts_with = f"{APPNAME}-"
    if only_system_folders:
        svc_ends_with = f"system-{instance_type.value}"
    elif only_user_folders:
        svc_ends_with = f"user-{instance_type.value}"
    else:
        svc_ends_with = f"-{instance_type.value}"

    matching_services = []
    for release in releases:
        release_name = release.get("name", "")

        # Check if this is a vantage6 service
        is_matching_service = (
            release_name.startswith(svc_starts_with)
            and release_name.endswith(svc_ends_with)
            and instance_type.value in release_name
        )

        if is_matching_service:
            matching_services.append(release_name)

    return matching_services


def select_running_service(
    running_services: list[str],
    instance_type: InstanceType,
) -> str:
    """
    Select a running service from the list of running services.
    """
    try:
        name = q.select(
            f"Select a {instance_type.value}:",
            choices=running_services,
        ).unsafe_ask()
    except KeyboardInterrupt:
        error("Aborted by user!")
        exit(1)
    return name


def get_main_cli_command_name(instance_type: InstanceType) -> str:
    """
    Get the main CLI command name for a given instance type.

    Parameters
    ----------
    instance_type : InstanceType
        The type of instance to get the main CLI command name for
    """
    if instance_type == InstanceType.HQ:
        return CLICommandName.HQ.value
    elif instance_type == InstanceType.ALGORITHM_STORE:
        return CLICommandName.ALGORITHM_STORE.value
    elif instance_type == InstanceType.NODE:
        return CLICommandName.NODE.value
    elif instance_type == InstanceType.AUTH:
        return CLICommandName.AUTH.value
    elif instance_type == InstanceType.HUB:
        return CLICommandName.HUB.value
    else:
        raise ValueError(f"Invalid instance type: {instance_type}")


def check_running(
    helm_release_name: str, instance_type: InstanceType, name: str, system_folders: bool
) -> bool:
    """
    Check if the instance is already running.

    Parameters
    ----------
    helm_release_name : str
        The name of the Helm release.
    instance_type : InstanceType
        The type of instance to check
    name : str
        The name of the instance to check
    system_folders : bool
        Whether to use system folders or not

    Returns
    -------
    bool
        True if the instance is already running, False otherwise
    """
    running_services = find_running_service_names(
        instance_type=instance_type,
        only_system_folders=system_folders,
        only_user_folders=not system_folders,
    )
    return helm_release_name in running_services


def get_config_name_from_helm_release_name(
    helm_release_name: str, is_store: bool = False
) -> str:
    """
    Get the config name from a helm release name.

    Parameters
    ----------
    helm_release_name : str
        The name of the Helm release
    is_store : bool, optional
        Whether the instance is a store or not. By default False.

    Returns
    -------
    str
        The config name
    """
    # helm release name is structured as:
    # f"{APPNAME}-{name}-{scope}-{instance_type}"
    # we want to get the name from the service name
    if is_store:
        # for store, the instance type is `algorithm-store` which contains an additional
        # hyphen
        return "-".join(helm_release_name.split("-")[1:-3])
    else:
        return "-".join(helm_release_name.split("-")[1:-2])


def extract_name_and_is_sandbox(name: str | None, is_sandbox: bool) -> tuple[str, bool]:
    """
    Extract the name and is_sandbox from the name.

    Note that the name may be None: this occurs before when this function is called
    before the user has selected a name. This scenario is fine because when the user
    selects a name interactively, the name never ends with the .sandbox suffix.

    Parameters
    ----------
    name : str | None
        The name of the instance
    is_sandbox : bool
        Whether the instance is a sandbox instance

    Returns
    -------
    tuple[str, bool]
        The name and is_sandbox
    """
    if name and name.endswith(SANDBOX_SUFFIX):
        return name[: -len(SANDBOX_SUFFIX)], True
    else:
        return name, is_sandbox


def generate_password(password_length: int = 16) -> str:
    """
    Generate a strong password that meets the password policy requirements.

    This ensures that the password has at least 8 characters, one uppercase letter,
    one lowercase letter, one number and one special character. We are avoiding
    quotes in the password to avoid issues with escaping in YAML.

    Parameters
    ----------
    password_length : int, optional
        The length of the password to generate. By default 16.

    Returns
    -------
    str
        The generated password
    """
    alphabet = string.ascii_letters + string.digits + string.punctuation
    # Avoid quotes to prevent YAML escaping issues
    alphabet = "".join(set(alphabet) - {"'", '"'})
    while True:
        password = "".join(secrets.choice(alphabet) for i in range(password_length))
        if (
            any(c.islower() for c in password)
            and any(c.isupper() for c in password)
            and any(c.isdigit() for c in password)
            and any(c in string.punctuation for c in password)
        ):
            break
    return password


def create_kubernetes_secret(
    core_api: k8s_client.CoreV1Api,
    secret_name: str,
    namespace: str,
    secret_data: dict[str, str],
) -> None:
    """
    Create a Kubernetes secret.

    Parameters
    ----------
    core_api: k8s_client.CoreV1Api
        The Kubernetes Core API instance
    secret_name: str
        The name of the secret
    namespace: str
        The namespace where the secret will be created
    secret_data: dict[str, str]
        The data to be stored in the secret
    """
    secret_body = k8s_client.V1Secret(
        metadata=k8s_client.V1ObjectMeta(name=secret_name, namespace=namespace),
        type="Opaque",
        string_data=secret_data,
    )
    core_api.create_namespaced_secret(namespace=namespace, body=secret_body)



import json
import subprocess
import sys
from pathlib import Path

import click
from rich.console import Console

from vantage6.cli.sandbox.new import cli_new_sandbox
from vantage6.cli.sandbox.remove import cli_sandbox_remove
from vantage6.cli.sandbox.start import cli_sandbox_start
from vantage6.cli.sandbox.stop import cli_sandbox_stop
from vantage6.cli.utils import prompt_config_name

TEST_FILE_PATH = Path(__file__).parent / "algo_test_scripts" / "algo_test_script.py"


@click.command()
@click.option(
    "--script",
    type=click.Path(),
    # default=TEST_FILE_PATH,
    help="Path of the script to test the algorithm.",
)
@click.option(
    "--task-arguments",
    type=str,
    default=None,
    help="Arguments to be provided to Task.create function. If --script is provided, this should not be set.",
)
@click.option(
    "--create-sandbox",
    is_flag=True,
    help="Create a new sandbox to run the test",
)
@click.option(
    "--start-sandbox",
    is_flag=True,
    help="Start a sandbox to run the test",
)
@click.option("-n", "--name", default=None, type=str, help="Name for your sandbox")
@click.option("--hq-image", type=str, default=None, help="HQ Docker image to use")
@click.option(
    "--store-image", type=str, default=None, help="Algorithm store Docker image to use"
)
@click.option("--node-image", type=str, default=None, help="Node Docker image to use")
@click.option(
    "--extra-hq-config",
    type=click.Path(exists=True),
    default=None,
    help="YAML File with additional HQ configuration. This will be appended to the HQ "
    "configuration file",
)
@click.option(
    "--extra-store-config",
    type=click.Path(exists=True),
    default=None,
    help="YAML File with additional algorithm store configuration. This will be "
    "appended to the algorithm store configuration file",
)
@click.option(
    "--extra-auth-config",
    type=click.Path(exists=True),
    default=None,
    help="YAML File with additional auth configuration. This will be appended to the "
    "auth configuration file",
)
@click.option(
    "--extra-node-config",
    type=click.Path(exists=True),
    default=None,
    help="YAML File with additional node configuration. This will be appended to the "
    "node configuration file",
)
@click.option(
    "--keep",
    type=bool,
    flag_value=True,
    default=False,
    help="Keep the sandbox after finishing the test",
)
@click.option(
    "--add-dataset",
    type=(str, click.Path()),
    default=[],
    multiple=True,
    help="Add a dataset to the nodes. The first argument is the label of the database, "
    "the second is the path to the dataset file.",
)
@click.pass_context
def cli_test_client_script(
    click_ctx: click.Context,
    script: Path | None,
    task_arguments: str | None,
    name: str,
    create_sandbox: bool,
    start_sandbox: bool,
    hq_image: str,
    store_image: str,
    node_image: str,
    extra_hq_config: Path,
    extra_store_config: Path,
    extra_auth_config: Path,
    extra_node_config: Path,
    keep: bool,
    add_dataset: list[tuple[str, Path]] = (),
) -> int:
    """
    Run a script for testing an algorithm on a sandbox network.
    The path to the script must be provided as an argument.
    """
    if not (script or task_arguments):
        raise click.UsageError("--script or --task-arguments must be set.")
    elif script and task_arguments:
        raise click.UsageError("--script and --task-arguments cannot be set together.")

    # Check if the task_arguments is a valid JSON string
    if task_arguments:
        try:
            json.loads(task_arguments.replace("'", '"'))
        except json.JSONDecodeError:
            raise click.UsageError("task-arguments must be a valid JSON string.")

    name = prompt_config_name(name)

    # create the network
    if create_sandbox:
        click_ctx.invoke(
            cli_new_sandbox,
            name=name,
            num_nodes=3,
            hq_image=hq_image,
            store_image=store_image,
            node_image=node_image,
            extra_hq_config=extra_hq_config,
            extra_store_config=extra_store_config,
            extra_auth_config=extra_auth_config,
            extra_node_config=extra_node_config,
            add_dataset=add_dataset,
        )

    # start the server and nodes
    if start_sandbox and not create_sandbox:
        click_ctx.invoke(
            cli_sandbox_start,
            name=name,
            node_image=node_image,
            extra_node_config=extra_node_config,
            add_dataset=add_dataset,
        )

    # run the test script and get the result
    if not task_arguments:
        subprocess_args = ["python", script]
    else:
        subprocess_args = ["python", script, task_arguments]

    result = subprocess.run(subprocess_args, stdout=sys.stdout, stderr=sys.stderr)

    # check the exit code. If the test passed, it should be 0
    if result.returncode == 0:
        msg = ":heavy_check_mark: [green]Test passed[/green]"
    else:
        msg = ":x: [red]Test failed[/red]"

    console = Console()
    console.print(msg)

    # clean up the test resources. Keep the network if --keep is set, or if the network
    # was created for this test. If the network was started for this test, stop it but
    # do not remove it.
    if not keep:
        if start_sandbox or create_sandbox:
            click_ctx.invoke(cli_sandbox_stop, name=name)
        if create_sandbox:
            click_ctx.invoke(cli_sandbox_remove, name=name)

    return result.returncode



import click

from vantage6.common.globals import InstanceType

from vantage6.cli.common.list import get_configuration_list


@click.command()
def cli_hq_configuration_list() -> None:
    """
    Print the available HQ configurations.
    """
    get_configuration_list(InstanceType.HQ)



from pathlib import Path

import click

from vantage6.common import info, warning
from vantage6.common.globals import InstanceType

from vantage6.cli.auth.install import check_and_install_keycloak_operator
from vantage6.cli.auth.k8s_utils import wait_for_keycloak_ready
from vantage6.cli.common.decorator import click_insert_context
from vantage6.cli.common.k8s_utils import wait_for_pod_ready
from vantage6.cli.common.start import helm_install, prestart_checks
from vantage6.cli.context.hub import HubContext
from vantage6.cli.globals import ChartName
from vantage6.cli.hub.install import (
    cert_manager_seems_installed,
    check_and_install_cert_manager_crds,
)
from vantage6.cli.hub.utils.gateway import ensure_envoy_gateway
from vantage6.cli.k8s_config import select_k8s_config


@click.command()
@click.option("--context", default=None, help="Kubernetes context to use")
@click.option("--namespace", default=None, help="Kubernetes namespace to use")
@click.option(
    "--local-chart-dir",
    type=click.Path(exists=True),
    default=None,
    help="Local chart repository to use.",
)
@click.option("--chart-version", default=None, help="Chart version to use")
@click.option("--sandbox/--no-sandbox", "sandbox", default=False)
@click.option("--wait-ready/--no-wait-ready", "wait_ready", default=True)
@click.option(
    "--auto-install-gateway/--no-auto-install-gateway",
    "auto_install_gateway",
    default=True,
    help="Automatically install Envoy Gateway if it is not detected in the cluster.",
)
@click.option(
    "--envoy-gateway-version",
    default=None,
    help=(
        "Envoy Gateway version tag to use for installation. "
        "If omitted, the latest release is looked up from GitHub; "
        "if that lookup fails, provide this flag explicitly."
    ),
)
@click_insert_context(
    type_=InstanceType.HUB,
    include_name=True,
    include_system_folders=True,
    sandbox_param="sandbox",
)
def cli_hub_start(
    ctx: HubContext,
    name: str,
    system_folders: bool,
    context: str | None,
    namespace: str | None,
    local_chart_dir: Path | None,
    chart_version: str | None,
    wait_ready: bool,
    auto_install_gateway: bool = True,
    envoy_gateway_version: str | None = None,
) -> None:
    """
    Start a hub environment.
    """
    info("Starting hub...")

    prestart_checks(ctx, InstanceType.HUB, name, system_folders)

    k8s_config = select_k8s_config(context=context, namespace=namespace)

    # Before starting the hub, ensure required operators / CRDs are installed.
    hub_gateway = ctx.config.get("hubGateway", {})
    tls_cfg = hub_gateway.get("tls", {})

    # 1) cert-manager CRDs are needed for Certificate resources used by the hub
    #    chart, but only when ingress/gateway exposure is enabled and configured
    #    to use cert-manager.
    if hub_gateway.get("enabled") and tls_cfg.get("mode") == "cert-manager":
        check_and_install_cert_manager_crds(k8s_config)
        if not cert_manager_seems_installed(k8s_config):
            warning(
                "⚠️  hubGateway.tls.mode is set to 'cert-manager', but cert-manager "
                "does not appear to be installed. No automatic installation will "
                "be attempted; please install and configure cert-manager (or "
                "switch hubGateway.tls.mode to 'existingSecret')."
            )

    # 2) Ensure an Envoy Gateway installation is available when hub gateway is enabled.
    if hub_gateway.get("enabled"):
        ensure_envoy_gateway(
            k8s_config,
            auto_install=auto_install_gateway,
            version=envoy_gateway_version,
        )

    # 3) Keycloak operator (and its CRDs) are needed for the auth subchart.
    check_and_install_keycloak_operator(k8s_config)

    helm_install(
        release_name=ctx.helm_release_name,
        chart_name=ChartName.HUB,
        values_file=ctx.config_file,
        k8s_config=k8s_config,
        local_chart_dir=local_chart_dir,
        chart_version=chart_version,
    )

    if wait_ready:
        info("Waiting for hub services to become ready (auth, hq, store)...")
        wait_for_keycloak_ready(ctx.helm_release_name, k8s_config)
        wait_for_pod_ready(
            selector=(
                f"app=vantage6-store,release={ctx.helm_release_name},component=store"
            ),
            description="Store",
            k8s_config=k8s_config,
        )
        wait_for_pod_ready(
            selector=(
                f"app=vantage6-hq,release={ctx.helm_release_name},component=vantage6-hq"
            ),
            description="HQ",
            k8s_config=k8s_config,
        )
        info("Hub services are ready.")
    else:
        info("Hub services have been started.")
        info("You may need to wait for the services to become ready.")



import os
import subprocess
import sys
import time

import click
import requests

from vantage6.common import error, info, warning

from vantage6.cli.common.k8s_utils import run_kubectl_command
from vantage6.cli.k8s_config import KubernetesConfig, select_k8s_config


def _get_latest_keycloak_version() -> str:
    """
    Get the latest Keycloak Operator version from GitHub releases. Exits when unable
    to find what version is the latest version.

    Returns
    -------
    str
        The latest version found in the Keycloak Github releases.
    """
    github_api_url = "https://api.github.com/repos/keycloak/keycloak/releases/latest"

    # Prepare headers with optional GitHub token for authentication
    headers = {"Accept": "application/vnd.github.v3+json"}
    github_token = os.environ.get("GITHUB_TOKEN")
    if github_token:
        headers["Authorization"] = f"token {github_token}"

    try:
        response = requests.get(
            github_api_url,
            timeout=5,
            headers=headers,
        )
        response.raise_for_status()
        release_data = response.json()
        version = release_data.get("tag_name")
        if not version:
            error("No version found in the Keycloak Github releases.")
            info("Please specify a version manually using the --operator-version flag.")
            exit(1)
        # Remove 'v' prefix if present (e.g., "v24.0.0" -> "24.0.0")
        if version.startswith("v"):
            version = version[1:]
        return version
    except requests.HTTPError as e:
        if e.response.status_code == 403:
            error(
                "GitHub API rate limit exceeded. "
                "Set GITHUB_TOKEN environment variable to authenticate and increase rate "
                "limits, or specify a version manually using the --operator-version flag."
            )
        else:
            error(
                f"Failed to fetch latest Keycloak Operator version from GitHub: {e}. "
            )
        info("Please specify a version manually using the --operator-version flag.")
        exit(1)
    except (requests.RequestException, KeyError, ValueError) as e:
        error(f"Failed to fetch latest Keycloak Operator version from GitHub: {e}. ")
        info("Please specify a version manually using the --operator-version flag.")
        exit(1)


@click.command()
@click.option("--context", default=None, help="Kubernetes context to use")
@click.option("--namespace", default=None, help="Kubernetes namespace for the operator")
@click.option(
    "--operator-version",
    default=None,
    help="Keycloak Operator version to install (defaults to latest from GitHub)",
)
@click.option(
    "--skip-crds/--no-skip-crds",
    default=False,
    help="Skip installing custom resource definitions (CRDs)",
)
def cli_auth_install_operator(
    context: str | None,
    namespace: str | None,
    operator_version: str | None,
    skip_crds: bool,
) -> None:
    """
    Install the Keycloak Operator and its CRDs in the Kubernetes cluster.

    This command installs the Keycloak Operator which is required to manage
    Keycloak instances using Custom Resources. This installation is done according to
    the docs in
    https://www.keycloak.org/operator/installation#_installing_by_using_kubectl_without_operator_lifecycle_manager
    """
    # Get the latest version if not specified
    if operator_version is None:
        operator_version = _get_latest_keycloak_version()

    k8s_config = select_k8s_config(context=context, namespace=namespace)

    # Base URL for Keycloak Operator manifests
    base_url = (
        "https://raw.githubusercontent.com/keycloak/keycloak-k8s-resources/"
        f"{operator_version}/kubernetes/"
    )

    # Install custom resource definitions (CRDs)
    crd_files_to_install = [
        "keycloaks.k8s.keycloak.org-v1.yml",
        "keycloakrealmimports.k8s.keycloak.org-v1.yml",
        "kubernetes.yml",
    ]
    if not skip_crds:
        info("Installing Keycloak CRDs...")
        for crd_file in crd_files_to_install:
            crd_url = f"{base_url}/{crd_file}"
            run_kubectl_command(
                ["apply", "-f", crd_url],
                k8s_config,
            )
        info("CRDs installed successfully.")

    # Adapt the clusterrolebinding to use the correct namespace (see
    # https://www.keycloak.org/operator/installation)
    run_kubectl_command(
        [
            "patch",
            "clusterrolebinding",
            "keycloak-operator-clusterrole-binding",
            "--type=json",
            '-p=[{"op": "replace", "path": "/subjects/0/namespace", "value":"'
            'custom-namespace"}]',
        ],
        k8s_config=k8s_config,
    )
    info("Clusterrolebinding adapted to use the correct namespace.")

    run_kubectl_command(
        [
            "rollout",
            "restart",
            "Deployment/keycloak-operator",
        ],
        k8s_config=k8s_config,
    )
    info("Keycloak Operator rolled out successfully.")

    # Wait for operator to be ready
    info("Waiting for Keycloak Operator to become ready...")
    _wait_for_operator_ready(k8s_config.namespace, k8s_config, timeout=300)

    info(
        "Keycloak Operator installed successfully in namespace "
        f"'{k8s_config.namespace}'."
    )


def _wait_for_operator_ready(
    namespace: str,
    k8s_config,
    timeout: int = 300,
) -> None:
    """
    Wait for the Keycloak Operator deployment to become ready.

    Parameters
    ----------
    namespace : str
        The namespace where the operator is installed.
    k8s_config
        Kubernetes configuration object.
    timeout : int
        Maximum time to wait in seconds (default: 300).
    """
    start_time = time.time()
    deployment_name = "keycloak-operator"

    while (time.time() - start_time) < timeout:
        try:
            command = [
                "kubectl",
                "get",
                "deployment",
                deployment_name,
                "-n",
                namespace,
                "-o",
                'jsonpath={.status.conditions[?(@.type=="Available")].status}',
            ]

            if k8s_config.context:
                command.extend(["--context", k8s_config.context])

            result = subprocess.run(
                command,
                capture_output=True,
                text=True,
                check=False,
            )

            if result.stdout.strip() == "True":
                info("Keycloak Operator is ready.")
                return

        except Exception as e:
            warning(f"Error checking operator status: {e}")

        time.sleep(5)

    error(f"Timeout: Keycloak Operator did not become ready within {timeout} seconds.")
    warning(
        "You can check the operator status manually with: "
        f"kubectl get deployment keycloak-operator -n {namespace}"
    )


def check_and_install_keycloak_operator(k8s_config: KubernetesConfig):
    try:
        k8s_config = select_k8s_config(context=None, namespace=None)
        if not _check_keycloak_operator_installed(k8s_config):
            warning("⚠️  Keycloak operator is not installed.")
            info("Installing Keycloak operator...")
            cmd = ["v6", "auth", "install-keycloak"]
            if k8s_config.context:
                cmd.extend(["--context", k8s_config.context])
            if k8s_config.namespace:
                cmd.extend(["--namespace", k8s_config.namespace])

            result = subprocess.run(cmd, check=False)
            if result.returncode != 0:
                error("❌ Failed to install Keycloak operator.")
                error("Please run 'v6 auth install-keycloak' manually and try again.")
                sys.exit(1)
            info("✅ Keycloak operator installed successfully.")
        else:
            info("✅ Keycloak operator is already installed.")
    except Exception as e:
        warning(f"⚠️  Could not check Keycloak operator status: {e}")
        warning(
            "Continuing anyway. If Keycloak fails to start, run 'v6 auth "
            "install-keycloak' manually."
        )


def _check_keycloak_operator_installed(k8s_config: KubernetesConfig) -> bool:
    """
    Check if the Keycloak operator is already installed.

    Parameters
    ----------
    k8s_config
        Kubernetes configuration object.

    Returns
    -------
    bool
        True if the operator is installed, False otherwise.
    """
    try:
        cmd = ["kubectl", "get", "deployment", "keycloak-operator"]
        if k8s_config.context:
            cmd.extend(["--context", k8s_config.context])
        if k8s_config.namespace:
            cmd.extend(["--namespace", k8s_config.namespace])

        result = subprocess.run(
            cmd,
            capture_output=True,
            text=True,
            check=False,
        )
        return result.returncode == 0
    except Exception:
        return False
